import typing

from telegrinder.bot.cute_types.chat_member_updated import ChatMemberUpdatedCute
from telegrinder.bot.dispatch.view.base import BaseStateView
from telegrinder.types.enums import UpdateType
from telegrinder.types.objects import Update

ChatMemberUpdateType: typing.TypeAlias = typing.Literal[
    UpdateType.CHAT_MEMBER,
    UpdateType.MY_CHAT_MEMBER,
]


class ChatMemberView(BaseStateView[ChatMemberUpdatedCute]):
    def __init__(self, *, update_type: ChatMemberUpdateType | None = None) -> None:
        self.auto_rules = []
        self.handlers = []
        self.middlewares = []
        self.return_manager = None
        self.update_type = update_type

    def __repr__(self) -> str:
        return "<{}: {!r}>".format(
            self.__class__.__name__,
            "chat_member_updated" if self.update_type is None else self.update_type.value,
        )

    def get_state_key(self, event: ChatMemberUpdatedCute) -> int | None:
        return event.chat_id

    async def check(self, event: Update) -> bool:
        return not (
            self.update_type is not None
            and self.update_type != event.update_type
            or not await super().check(event)
        )


__all__ = ("ChatMemberView",)



import typing

import vbml

from telegrinder.bot.dispatch.context import Context
from telegrinder.node.text import Text
from telegrinder.tools.global_context.telegrinder_ctx import TelegrinderContext

from .abc import ABCRule

PatternLike: typing.TypeAlias = str | vbml.Pattern
global_ctx: typing.Final[TelegrinderContext] = TelegrinderContext()


def check_string(patterns: list[vbml.Pattern], s: str, ctx: Context) -> bool:
    for pattern in patterns:
        match global_ctx.vbml_patcher.check(pattern, s):
            case None | False:
                continue
            case {**response}:
                ctx |= response
        return True
    return False


class Markup(ABCRule):
    """Markup Language. See [VBML docs](https://github.com/tesseradecade/vbml/blob/master/docs/index.md)"""

    def __init__(self, patterns: PatternLike | list[PatternLike], /) -> None:
        if not isinstance(patterns, list):
            patterns = [patterns]
        self.patterns = [
            vbml.Pattern(pattern) if isinstance(pattern, str) else pattern for pattern in patterns
        ]

    async def check(self, text: Text, ctx: Context) -> bool:
        return check_string(self.patterns, text, ctx)


__all__ = ("Markup", "check_string")



import dataclasses
import inspect
import typing
from functools import wraps

import typing_extensions
from fntypes.result import Result

from telegrinder.api.api import API
from telegrinder.model import Model, get_params

F = typing.TypeVar("F", bound=typing.Callable[..., typing.Any])
Cute = typing.TypeVar("Cute", bound="BaseCute")
Update = typing_extensions.TypeVar("Update", bound=Model)
CtxAPI = typing_extensions.TypeVar("CtxAPI", bound=API, default=API)

Executor: typing.TypeAlias = typing.Callable[
    [Cute, str, dict[str, typing.Any]],
    typing.Awaitable[Result[typing.Any, typing.Any]],
]

if typing.TYPE_CHECKING:

    class BaseCute(Model, typing.Generic[Update, CtxAPI]):
        api: API

        @classmethod
        def from_update(cls, update: Update, bound_api: API) -> typing.Self: ...

        @property
        def ctx_api(self) -> CtxAPI: ...

        def to_dict(
            self,
            *,
            exclude_fields: set[str] | None = None,
        ) -> dict[str, typing.Any]:
            """
            :param exclude_fields: Cute model field names to exclude from the dictionary representation of this cute model.
            :return: A dictionary representation of this cute model.
            """

            ...

        def to_full_dict(
            self,
            *,
            exclude_fields: set[str] | None = None,
        ) -> dict[str, typing.Any]:
            """
            :param exclude_fields: Cute model field names to exclude from the dictionary representation of this cute model.
            :return: A dictionary representation of this model including all models, structs, custom types.
            """

            ...

else:
    from fntypes.co import Nothing, Some, Variative

    from telegrinder.msgspec_utils import Option, decoder
    from telegrinder.msgspec_utils import get_class_annotations as _get_class_annotations

    def _get_cute_from_generic(generic_args):
        for arg in generic_args:
            orig_arg = typing.get_origin(arg) or arg

            if not isinstance(orig_arg, type):
                continue
            if orig_arg in (Variative, Some, Option):
                return _get_cute_from_generic(typing.get_args(arg))
            if issubclass(arg, BaseCute):
                return arg

        return None

    def _get_cute_annotations(annotations):
        cute_annotations = {}

        for key, hint in annotations.items():
            if not isinstance(hint, type):
                if (cute := _get_cute_from_generic(typing.get_args(hint))) is not None:
                    cute_annotations[key] = cute

            elif issubclass(hint, BaseCute):
                cute_annotations[key] = hint

        return cute_annotations

    def _get_value(value):
        while isinstance(value, Variative | Some):
            if isinstance(value, Variative):
                value = value.v
            if isinstance(value, Some):
                value = value.value
        return value

    class BaseCute(typing.Generic[Update, CtxAPI]):
        def __init_subclass__(cls, *args, **kwargs):
            super().__init_subclass__(*args, **kwargs)

            if not cls.__bases__ or not issubclass(cls.__bases__[0], BaseCute):
                return

            cls.__is_solved_annotations__ = False
            cls.__cute_annotations__ = None

        @classmethod
        def from_update(cls, update, bound_api):
            if not cls.__is_solved_annotations__:
                cls.__is_solved_annotations__ = True
                cls.__annotations__ = _get_class_annotations(cls)

            if cls.__cute_annotations__ is None:
                cls.__cute_annotations__ = _get_cute_annotations(cls.__annotations__)

            return cls(
                **{
                    field: decoder.convert(
                        cls.__cute_annotations__[field].from_update(_get_value(value), bound_api=bound_api),
                        type=cls.__annotations__[field],
                    )
                    if field in cls.__cute_annotations__ and not isinstance(value, Nothing)
                    else value
                    for field, value in update.to_dict().items()
                },
                api=bound_api,
            )

        @property
        def ctx_api(self):
            return self.api

        def to_dict(self, *, exclude_fields=None):
            exclude_fields = exclude_fields or set()
            return super().to_dict(exclude_fields={"api"} | exclude_fields)

        def to_full_dict(self, *, exclude_fields=None):
            exclude_fields = exclude_fields or set()
            return super().to_full_dict(exclude_fields={"api"} | exclude_fields)


def compose_method_params(
    params: dict[str, typing.Any],
    update: Cute,
    *,
    default_params: set[str | tuple[str, str]] | None = None,
    validators: dict[str, typing.Callable[[Cute], bool]] | None = None,
) -> dict[str, typing.Any]:
    """Compose method `params` from `update` by `default_params` and `validators`.

    :param params: Method params.
    :param update: Update object.
    :param default_params: Default params. \
    (`str`) - Attribute name to be get from `update` if param is undefined. \
    (`tuple[str, str]`): tuple[0] - Parameter name to be set in `params`, \
    tuple[1] - attribute name to be get from `update`.
    :param validators: Validators mapping (`str, Callable`), key - `Parameter name` \
    for which the validator will be applied, value - `Validator`, if returned `True` \
    parameter will be set, otherwise will not.
    :return: Composed params.
    """

    default_params = default_params or set()
    validators = validators or {}

    for param in default_params:
        param_name = param if isinstance(param, str) else param[0]
        if param_name not in params:
            if param_name in validators and not validators[param_name](update):
                continue
            params[param_name] = getattr(update, param if isinstance(param, str) else param[1])

    return params


def shortcut(
    method_name: str,
    *,
    executor: Executor[Cute] | None = None,
    custom_params: set[str] | None = None,
):
    def wrapper(func: F) -> F:
        @wraps(func)
        async def inner(
            self: Cute,
            *args: typing.Any,
            **kwargs: typing.Any,
        ) -> typing.Any:
            if executor is None:
                return await func(self, *args, **kwargs)

            if not hasattr(func, "_signature_params"):
                setattr(
                    func,
                    "_signature_params",
                    {k: p for k, p in inspect.signature(func).parameters.items() if k != "self"},
                )

            signature_params: dict[str, inspect.Parameter] = getattr(func, "_signature_params")
            params: dict[str, typing.Any] = {}
            index = 0

            for k, p in signature_params.items():
                if p.kind in (p.POSITIONAL_OR_KEYWORD, p.POSITIONAL_ONLY) and len(args) > index:
                    params[k] = args[index]
                    index += 1
                    continue

                if p.kind in (p.VAR_KEYWORD, p.VAR_POSITIONAL):
                    params[k] = kwargs.copy() if p.kind is p.VAR_KEYWORD else args[index:]
                    continue

                params[k] = kwargs.pop(k, p.default) if p.default is not p.empty else kwargs.pop(k)

            return await executor(self, method_name, get_params(params))

        inner.__shortcut__ = Shortcut(  # type: ignore
            method_name=method_name,
            executor=executor,
            custom_params=custom_params or set(),
        )
        return inner  # type: ignore

    return wrapper


@dataclasses.dataclass(slots=True, frozen=True)
class Shortcut:
    method_name: str
    executor: Executor | None = dataclasses.field(default=None, kw_only=True)
    custom_params: set[str] = dataclasses.field(default_factory=lambda: set(), kw_only=True)


__all__ = ("BaseCute", "Shortcut", "compose_method_params", "shortcut")



import difflib

from telegrinder.bot.dispatch.context import Context
from telegrinder.node.text import Text

from .abc import ABCRule


class FuzzyText(ABCRule):
    def __init__(self, texts: str | list[str], min_ratio: float = 0.7) -> None:
        if isinstance(texts, str):
            texts = [texts]
        self.texts = texts
        self.min_ratio = min_ratio

    async def check(self, message_text: Text, ctx: Context) -> bool:
        match = max(difflib.SequenceMatcher(a=message_text, b=text).ratio() for text in self.texts)
        if match < self.min_ratio:
            return False
        ctx.fuzzy_ratio = match
        return True


__all__ = ("FuzzyText",)



from fntypes.result import Ok, Result

from telegrinder.api import API
from telegrinder.bot.cute_types.update import UpdateCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.adapter.abc import ABCAdapter
from telegrinder.bot.rules.adapter.errors import AdapterError
from telegrinder.types.objects import Update


class RawUpdateAdapter(ABCAdapter[Update, UpdateCute]):
    ADAPTED_VALUE_KEY: str = "_adapted_update_cute"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: adapt Update -> UpdateCute>"

    async def adapt(
        self,
        api: API,
        update: Update,
        context: Context,
    ) -> Result[UpdateCute, AdapterError]:
        if self.ADAPTED_VALUE_KEY not in context:
            context[self.ADAPTED_VALUE_KEY] = (
                UpdateCute.from_update(update, api)
                if not isinstance(update, UpdateCute)
                else update
            )
        return Ok(context[self.ADAPTED_VALUE_KEY])


__all__ = ("RawUpdateAdapter",)



import abc
import inspect
import typing
from contextlib import suppress

import msgspec

from telegrinder.bot.cute_types import CallbackQueryCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.adapter import EventAdapter
from telegrinder.model import decoder
from telegrinder.tools.buttons import DataclassInstance
from telegrinder.types.enums import UpdateType

from .abc import ABCRule
from .markup import Markup, PatternLike, check_string

CallbackQuery: typing.TypeAlias = CallbackQueryCute
Validator: typing.TypeAlias = typing.Callable[[typing.Any], bool | typing.Awaitable[bool]]
MapDict: typing.TypeAlias = dict[str, "typing.Any | type[typing.Any] | Validator | list[MapDict] | MapDict"]
CallbackMap: typing.TypeAlias = list[tuple[str, "typing.Any | type[typing.Any] | Validator | CallbackMap"]]
CallbackMapStrict: typing.TypeAlias = list[tuple[str, "Validator | CallbackMapStrict"]]


class CallbackQueryRule(ABCRule[CallbackQuery], abc.ABC):
    adapter: EventAdapter[CallbackQuery] = EventAdapter(UpdateType.CALLBACK_QUERY, CallbackQuery)

    @abc.abstractmethod
    async def check(self, event: CallbackQuery, context: Context) -> bool:
        pass


class HasData(CallbackQueryRule):
    async def check(self, event: CallbackQuery) -> bool:
        return bool(event.data.unwrap_or_none())


class CallbackQueryDataRule(CallbackQueryRule, abc.ABC, requires=[HasData()]):
    pass


class CallbackDataMap(CallbackQueryDataRule):
    def __init__(self, mapping: MapDict, /) -> None:
        self.mapping = self.transform_to_callbacks(
            self.transform_to_map(mapping),
        )

    @classmethod
    def transform_to_map(cls, mapping: MapDict) -> CallbackMap:
        """Transforms MapDict to CallbackMap."""

        callback_map = []

        for k, v in mapping.items():
            if isinstance(v, dict):
                v = cls.transform_to_map(v)
            callback_map.append((k, v))

        return callback_map

    @classmethod
    def transform_to_callbacks(cls, callback_map: CallbackMap) -> CallbackMapStrict:
        """Transforms `CallbackMap` to `CallbackMapStrict`."""

        callback_map_result = []

        for key, value in callback_map:
            if isinstance(value, type):
                validator = (lambda tp: lambda v: isinstance(v, tp))(value)
            elif isinstance(value, list):
                validator = cls.transform_to_callbacks(value)
            elif not callable(value):
                validator = (lambda val: lambda v: val == v)(value)
            else:
                validator = value
            callback_map_result.append((key, validator))

        return callback_map_result

    @staticmethod
    async def run_validator(value: typing.Any, validator: Validator) -> bool:
        """Run async or sync validator."""

        with suppress(BaseException):
            result = validator(value)
            if inspect.isawaitable(result):
                result = await result
            return result

        return False

    @classmethod
    async def match(cls, callback_data: dict[str, typing.Any], callback_map: CallbackMapStrict) -> bool:
        """Matches callback_data with callback_map recursively."""

        for key, validator in callback_map:
            if key not in callback_data:
                return False

            if isinstance(validator, list):
                if not (
                    isinstance(callback_data[key], dict) and await cls.match(callback_data[key], validator)
                ):
                    return False

            elif not await cls.run_validator(callback_data[key], validator):
                return False

        return True

    async def check(self, event: CallbackQuery, ctx: Context) -> bool:
        callback_data = event.decode_callback_data().unwrap_or_none()
        if callback_data is None:
            return False
        if await self.match(callback_data, self.mapping):
            ctx.update(callback_data)
            return True
        return False


class CallbackDataEq(CallbackQueryDataRule):
    def __init__(self, value: str, /) -> None:
        self.value = value

    async def check(self, event: CallbackQuery) -> bool:
        return event.data.unwrap() == self.value


class CallbackDataJsonEq(CallbackQueryDataRule):
    def __init__(self, d: dict[str, typing.Any], /) -> None:
        self.d = d

    async def check(self, event: CallbackQuery) -> bool:
        return event.decode_callback_data().unwrap_or_none() == self.d


class CallbackDataJsonModel(CallbackQueryDataRule):
    def __init__(
        self,
        model: type[msgspec.Struct] | type[DataclassInstance],
        *,
        alias: str | None = None,
    ) -> None:
        self.model = model
        self.alias = alias or "data"

    async def check(self, event: CallbackQuery, ctx: Context) -> bool:
        with suppress(BaseException):
            ctx.set(self.alias, decoder.decode(event.data.unwrap().encode(), type=self.model))
            return True
        return False


class CallbackDataMarkup(CallbackQueryDataRule):
    def __init__(self, patterns: PatternLike | list[PatternLike], /) -> None:
        self.patterns = Markup(patterns).patterns

    async def check(self, event: CallbackQuery, ctx: Context) -> bool:
        return check_string(self.patterns, event.data.unwrap(), ctx)


__all__ = (
    "CallbackDataEq",
    "CallbackDataJsonEq",
    "CallbackDataJsonModel",
    "CallbackDataMap",
    "CallbackDataMarkup",
    "CallbackQueryDataRule",
    "CallbackQueryRule",
    "HasData",
)



from telegrinder.bot.dispatch.waiter_machine.hasher import (
    CALLBACK_QUERY_FOR_MESSAGE,
    CALLBACK_QUERY_FROM_CHAT,
    CALLBACK_QUERY_IN_CHAT_FOR_MESSAGE,
    MESSAGE_FROM_USER,
    MESSAGE_FROM_USER_IN_CHAT,
    MESSAGE_IN_CHAT,
    Hasher,
    StateViewHasher,
)
from telegrinder.bot.dispatch.waiter_machine.machine import WaiterMachine, clear_wm_storage_worker
from telegrinder.bot.dispatch.waiter_machine.middleware import WaiterMiddleware
from telegrinder.bot.dispatch.waiter_machine.short_state import ShortState

__all__ = (
    "CALLBACK_QUERY_FOR_MESSAGE",
    "CALLBACK_QUERY_FROM_CHAT",
    "CALLBACK_QUERY_IN_CHAT_FOR_MESSAGE",
    "Hasher",
    "MESSAGE_FROM_USER",
    "MESSAGE_FROM_USER_IN_CHAT",
    "MESSAGE_IN_CHAT",
    "ShortState",
    "StateViewHasher",
    "WaiterMachine",
    "WaiterMiddleware",
    "clear_wm_storage_worker",
)



import typing

from telegrinder.bot.cute_types.callback_query import CallbackQueryCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.return_manager.abc import BaseReturnManager, register_manager


class CallbackQueryReturnManager(BaseReturnManager[CallbackQueryCute]):
    @register_manager(str)
    @staticmethod
    async def str_manager(value: str, event: CallbackQueryCute, ctx: Context) -> None:
        await event.answer(value)

    @register_manager(dict[str, typing.Any])
    @staticmethod
    async def dict_manager(value: dict[str, typing.Any], event: CallbackQueryCute, ctx: Context) -> None:
        await event.answer(**value)


__all__ = ("CallbackQueryReturnManager",)



from telegrinder.bot.cute_types.update import UpdateCute
from telegrinder.types.enums import UpdateType

from .abc import ABCRule


class IsUpdateType(ABCRule):
    def __init__(self, update_type: UpdateType, /) -> None:
        self.update_type = update_type

    async def check(self, event: UpdateCute) -> bool:
        return event.update_type == self.update_type


__all__ = ("IsUpdateType",)



from telegrinder.bot.dispatch.return_manager.abc import (
    ABCReturnManager,
    BaseReturnManager,
    Manager,
    register_manager,
)
from telegrinder.bot.dispatch.return_manager.callback_query import CallbackQueryReturnManager
from telegrinder.bot.dispatch.return_manager.inline_query import InlineQueryReturnManager
from telegrinder.bot.dispatch.return_manager.message import MessageReturnManager

__all__ = (
    "ABCReturnManager",
    "BaseReturnManager",
    "CallbackQueryReturnManager",
    "InlineQueryReturnManager",
    "Manager",
    "MessageReturnManager",
    "register_manager",
)



import typing
from dataclasses import dataclass, field

from fntypes.option import Nothing, Option, Some

from telegrinder.node.base import DataNode
from telegrinder.node.text import Text


def single_split(s: str, separator: str) -> tuple[str, str]:
    left, *right = s.split(separator, 1)
    return left, (right[0] if right else "")


def cut_mention(text: str) -> tuple[str, Option[str]]:
    left, right = single_split(text, "@")
    return left, Some(right) if right else Nothing()


@dataclass(slots=True)
class CommandInfo(DataNode):
    name: str
    arguments: str
    mention: Option[str] = field(default_factory=Nothing)

    @classmethod
    async def compose(cls, text: Text) -> typing.Self:
        name, arguments = single_split(text, separator=" ")
        name, mention = cut_mention(name)
        return cls(name, arguments, mention)


__all__ = ("CommandInfo", "cut_mention", "single_split")



import inspect
from abc import ABC, abstractmethod
from functools import cached_property

import typing_extensions as typing

from telegrinder.bot.cute_types import MessageCute, UpdateCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.process import check_rule
from telegrinder.bot.rules.adapter import ABCAdapter, RawUpdateAdapter
from telegrinder.bot.rules.adapter.node import Event
from telegrinder.node.base import Node, get_nodes, is_node
from telegrinder.tools.i18n.base import ABCTranslator
from telegrinder.tools.magic import (
    cache_translation,
    get_annotations,
    get_cached_translation,
    get_default_args,
)
from telegrinder.types.objects import Update as UpdateObject

if typing.TYPE_CHECKING:
    from telegrinder.node.composer import NodeCollection

AdaptTo = typing.TypeVar("AdaptTo", default=typing.Any)

Message: typing.TypeAlias = MessageCute
Update: typing.TypeAlias = UpdateCute


def with_caching_translations(func: typing.Callable[..., typing.Any]):
    """Should be used as decorator for .translate method. Caches rule translations."""

    async def wrapper(self: "ABCRule", translator: ABCTranslator):
        if translation := get_cached_translation(self, translator.locale):
            return translation
        translation = await func(self, translator)
        cache_translation(self, translator.locale, translation)
        return translation

    return wrapper


class ABCRule(ABC, typing.Generic[AdaptTo]):
    adapter: ABCAdapter[UpdateObject, AdaptTo]
    requires: list["ABCRule"] = []

    if not typing.TYPE_CHECKING:
        adapter = RawUpdateAdapter()

    @abstractmethod
    async def check(self, event: AdaptTo, *, ctx: Context) -> bool:
        pass

    def __init_subclass__(cls, requires: list["ABCRule"] | None = None) -> None:
        """Merges requirements from inherited classes and rule-specific requirements."""

        requirements = []
        for base in inspect.getmro(cls):
            if issubclass(base, ABCRule) and base != cls:
                requirements.extend(base.requires or ())  # type: ignore

        requirements.extend(requires or ())
        cls.requires = list(dict.fromkeys(requirements))

    def __and__(self, other: "ABCRule") -> "AndRule":
        """And Rule.

        ```python
        rule = HasText() & HasCaption()
        rule #> AndRule(HasText(), HasCaption()) -> True if all rules in an AndRule are True, otherwise False.
        ```
        """

        return AndRule(self, other)

    def __or__(self, other: "ABCRule") -> "OrRule":
        """Or Rule.

        ```python
        rule = HasText() | HasCaption()
        rule #> OrRule(HasText(), HasCaption()) -> True if any rule in an OrRule are True, otherwise False.
        ```
        """

        return OrRule(self, other)

    def __invert__(self) -> "NotRule":
        """Not Rule.

        ```python
        rule = ~HasText()
        rule # NotRule(HasText()) -> True if rule returned False, otherwise False.
        ```
        """

        return NotRule(self)

    def __repr__(self) -> str:
        return "<{}: adapter={!r}>".format(
            self.__class__.__name__,
            self.adapter,
        )

    @cached_property
    def required_nodes(self) -> dict[str, type[Node]]:
        return get_nodes(self.check)

    def as_optional(self) -> "ABCRule":
        return self | Always()

    def should_fail(self) -> "ABCRule":
        return self & Never()

    async def bounding_check(
        self,
        adapted_value: AdaptTo,
        ctx: Context,
        node_col: "NodeCollection | None" = None,
    ) -> bool:
        bound_check_rule = self.check
        kw = {}
        node_col_values = node_col.values if node_col is not None else {}
        temp_ctx = get_default_args(bound_check_rule) | ctx

        for i, (k, v) in enumerate(get_annotations(bound_check_rule).items()):
            if (isinstance(adapted_value, Event) and i == 0) or (  # First arg is Event
                isinstance(v, type) and isinstance(adapted_value, v)
            ):
                kw[k] = adapted_value if not isinstance(adapted_value, Event) else adapted_value.obj
            elif is_node(v):
                assert k in node_col_values, "Node is undefined, error while bounding."
                kw[k] = node_col_values[k]
            elif k in temp_ctx:
                kw[k] = temp_ctx[k]
            elif v is Context:
                kw[k] = ctx
            else:
                raise LookupError(
                    f"Cannot bound {k!r} of type {v!r} to '{self.__class__.__qualname__}.check()', "
                    "because it cannot be resolved."
                )

        return await bound_check_rule(**kw)

    async def translate(self, translator: ABCTranslator) -> typing.Self:
        return self


class AndRule(ABCRule):
    def __init__(self, *rules: ABCRule) -> None:
        self.rules = rules

    async def check(self, event: Update, ctx: Context) -> bool:
        ctx_copy = ctx.copy()
        for rule in self.rules:
            if not await check_rule(event.ctx_api, rule, event, ctx_copy):
                return False
        ctx |= ctx_copy
        return True


class OrRule(ABCRule):
    def __init__(self, *rules: ABCRule) -> None:
        self.rules = rules

    async def check(self, event: Update, ctx: Context) -> bool:
        for rule in self.rules:
            ctx_copy = ctx.copy()
            if await check_rule(event.ctx_api, rule, event, ctx_copy):
                ctx |= ctx_copy
                return True
        return False


class NotRule(ABCRule):
    def __init__(self, rule: ABCRule) -> None:
        self.rule = rule

    async def check(self, event: Update, ctx: Context) -> bool:
        ctx_copy = ctx.copy()
        return not await check_rule(event.ctx_api, self.rule, event, ctx_copy)


class Never(ABCRule):
    async def check(self) -> typing.Literal[False]:
        return False


class Always(ABCRule):
    async def check(self) -> typing.Literal[True]:
        return True


__all__ = (
    "ABCRule",
    "Always",
    "AndRule",
    "Never",
    "NotRule",
    "OrRule",
    "with_caching_translations",
)



# NOTE: NEED REFACTORING

import html
import string
import typing
from contextlib import suppress

from telegrinder.tools.parse_mode import ParseMode
from telegrinder.types.enums import ProgrammingLanguage

from .links import (
    get_channel_boost_link,
    get_invite_chat_link,
    get_mention_link,
    get_resolve_domain_link,
    get_start_bot_link,
    get_start_group_link,
)
from .spec_html_formats import SpecialFormat, is_spec_format

TAG_FORMAT = "<{tag}{data}>{content}</{tag}>"
QUOT_MARK = '"'


class StringFormatterProto(typing.Protocol):
    def format_field(self, value: typing.Any, fmt: str) -> "HTMLFormatter": ...

    def format(self, __string: str, *args: object, **kwargs: object) -> "HTMLFormatter": ...


class StringFormatter(string.Formatter):
    """String formatter, using substitutions from args and kwargs.
    The substitutions are identified by braces ('{' and '}') with
    specifiers: `bold`, `italic`, etc.
    """

    __formats__: typing.ClassVar[tuple[str, ...]] = (
        "blockquote",
        "bold",
        "code_inline",
        "italic",
        "spoiler",
        "strike",
        "underline",
    )

    def is_html_format(self, value: typing.Any, fmt: str) -> str:
        if not fmt:
            raise ValueError("Formats union should be: format+format.")
        if fmt not in self.__formats__:
            raise ValueError(
                "Unknown format {!r} for object of type {!r}.".format(
                    fmt,
                    type(value).__name__,
                )
            )
        return fmt

    def get_spec_formatter(self, value: SpecialFormat) -> typing.Callable[..., "TagFormat"]:
        return globals()[value.__formatter_name__]

    def check_formats(self, value: typing.Any, fmts: list[str]) -> "TagFormat":
        if is_spec_format(value):
            value = value.string

        current_format = globals()[fmts.pop(0)](
            str(value)
            if isinstance(value, TagFormat)
            else escape(FormatString(value) if not isinstance(value, str) else value)
        )
        for fmt in fmts:
            current_format = globals()[fmt](current_format)

        return (
            TagFormat(
                current_format,
                tag=value.tag,
                **value.data,
            )
            if isinstance(value, TagFormat)
            else current_format
        )

    def format_field(self, value: typing.Any, fmt: str) -> "HTMLFormatter":
        with suppress(ValueError):
            return HTMLFormatter(
                format(
                    (
                        value.formatting()
                        if isinstance(value, TagFormat)
                        else (
                            self.get_spec_formatter(value)(**value.__dict__).formatting()
                            if is_spec_format(value)
                            else value
                        )
                    ),
                    fmt,
                )
            )
        return self.format_raw_value(value, fmt)

    def format_raw_value(self, value: typing.Any, fmt: str) -> "HTMLFormatter":
        fmts = list(map(lambda fmt: self.is_html_format(value, fmt), fmt.split("+")))
        tag_format = self.check_formats(value, fmts)

        if is_spec_format(value):
            value.string = tag_format
            tag_format = self.get_spec_formatter(value)(**value.__dict__)

        return tag_format.formatting()

    def format(self, __string: str, *args: object, **kwargs: object) -> "HTMLFormatter":
        return HTMLFormatter(super().format(__string, *args, **kwargs))


class FormatString(str):
    STRING_FORMATTER: StringFormatterProto = StringFormatter()

    def __new__(cls, string: str) -> typing.Self:
        if isinstance(string, TagFormat):
            return super().__new__(cls, string.formatting())
        return super().__new__(cls, string)

    def __add__(self, value: str) -> "HTMLFormatter":
        return HTMLFormatter(
            str.__add__(
                escape(self),
                value.formatting() if isinstance(value, TagFormat) else escape(value),
            )
        )

    def __radd__(self, value: str) -> "HTMLFormatter":
        """Return value+self."""
        return HTMLFormatter(FormatString.__add__(FormatString(value), self).as_str())

    def as_str(self) -> str:
        """Return self as a standart string."""
        return self.__str__()

    def format(self, *args: object, **kwargs: object) -> "HTMLFormatter":
        return self.STRING_FORMATTER.format(self, *args, **kwargs)


class EscapedString(FormatString):
    @property
    def former_string(self) -> str:
        return html.unescape(self)


class TagFormat(FormatString):
    tag: str
    data: dict[str, typing.Any]

    def __new__(
        cls,
        string: str,
        *,
        tag: str,
        **data: typing.Any,
    ) -> typing.Self:
        if isinstance(string, TagFormat):
            string = string.formatting()
        elif not isinstance(
            string,
            (
                EscapedString,
                HTMLFormatter,
            ),
        ):
            string = escape(string)
        obj = super().__new__(cls, string)
        obj.tag = tag
        obj.data = data
        return obj

    def get_tag_data(self) -> str:
        return "".join(f" {k}={v}" for k, v in self.data.items())

    def formatting(self) -> "HTMLFormatter":
        return HTMLFormatter(
            TAG_FORMAT.format(
                tag=self.tag,
                data=self.get_tag_data(),
                content=self,
            )
        )


class HTMLFormatter(FormatString):
    """
    >>> HTMLFormatter(bold("Hello, World"))
    '<b>Hello, World</b>'
    HTMLFormatter("Hi, {name:italic}").format(name="Max")
    'Hi, <i>Max</i>'
    """

    PARSE_MODE = ParseMode.HTML


def escape(string: str) -> EscapedString:
    if isinstance(string, EscapedString | HTMLFormatter):
        return EscapedString(string)
    return EscapedString(html.escape(string, quote=False))


def block_quote(string: str) -> TagFormat:
    return TagFormat(string, tag="blockquote")


def bold(string: str) -> TagFormat:
    return TagFormat(string, tag="b")


def channel_boost_link(channel_id: str | int, string: str | None = None):
    return link(get_channel_boost_link(channel_id), string)


def code_inline(string: str) -> TagFormat:
    return TagFormat(string, tag="code")


def italic(string: str) -> TagFormat:
    return TagFormat(string, tag="i")


def link(href: str, string: str | None = None) -> TagFormat:
    return TagFormat(
        string or href,
        tag="a",
        href=QUOT_MARK + href + QUOT_MARK,
    )


def pre_code(string: str, lang: str | ProgrammingLanguage | None = None) -> TagFormat:
    if lang is None:
        return TagFormat(string, tag="pre")
    lang = lang.value if isinstance(lang, ProgrammingLanguage) else lang
    return pre_code(TagFormat(string, tag="code", **{"class": f"language-{lang}"}))


def spoiler(string: str) -> TagFormat:
    return TagFormat(string, tag="tg-spoiler")


def start_bot_link(bot_id: str | int, data: str, string: str | None = None) -> TagFormat:
    return link(get_start_bot_link(bot_id, data), string)


def start_group_link(bot_id: str | int, data: str, string: str | None = None) -> TagFormat:
    return link(get_start_group_link(bot_id, data), string)


def strike(string: str) -> TagFormat:
    return TagFormat(string, tag="s")


def mention(string: str, user_id: int) -> TagFormat:
    return link(get_mention_link(user_id), string)


def tg_emoji(string: str, emoji_id: int) -> TagFormat:
    return TagFormat(string, tag="tg-emoji", **{"emoji-id": emoji_id})


def invite_chat_link(invite_link: str, string: str | None = None) -> TagFormat:
    return link(
        get_invite_chat_link(invite_link),
        string or f"https://t.me/joinchat/{invite_link}",
    )


def resolve_domain(username: str, string: str | None = None) -> TagFormat:
    return link(
        get_resolve_domain_link(username),
        string or f"t.me/{username}",
    )


def underline(string: str) -> TagFormat:
    return TagFormat(string, tag="u")


__all__ = (
    "FormatString",
    "HTMLFormatter",
    "SpecialFormat",
    "block_quote",
    "bold",
    "channel_boost_link",
    "code_inline",
    "escape",
    "get_channel_boost_link",
    "get_invite_chat_link",
    "get_mention_link",
    "get_resolve_domain_link",
    "get_start_bot_link",
    "get_start_group_link",
    "invite_chat_link",
    "italic",
    "link",
    "mention",
    "pre_code",
    "resolve_domain",
    "spoiler",
    "start_bot_link",
    "start_group_link",
    "strike",
    "tg_emoji",
    "underline",
)



import enum
import typing

from telegrinder.bot.dispatch.context import Context
from telegrinder.node.text import Text

from .abc import ABCRule

T = typing.TypeVar("T", bound=enum.Enum)


class EnumTextRule(ABCRule, typing.Generic[T]):
    def __init__(self, enum_t: type[T], *, lower_case: bool = True) -> None:
        self.enum_t = enum_t
        self.texts = list(
            map(
                lambda x: x.value.lower() if lower_case else x.value,
                self.enum_t,
            )
        )

    def find(self, s: str) -> T:
        for enumeration in self.enum_t:
            if enumeration.value.lower() == s:
                return enumeration
        raise KeyError("Enumeration is undefined.")

    async def check(self, text: Text, ctx: Context) -> bool:
        text = text.lower()  # type: ignore
        if text not in self.texts:
            return False
        ctx.enum_text = self.find(text)
        return True



from telegrinder.bot.cute_types.callback_query import CallbackQueryCute
from telegrinder.bot.dispatch.return_manager.callback_query import CallbackQueryReturnManager
from telegrinder.bot.dispatch.view.base import BaseStateView


class CallbackQueryView(BaseStateView[CallbackQueryCute]):
    def __init__(self) -> None:
        self.auto_rules = []
        self.handlers = []
        self.middlewares = []
        self.return_manager = CallbackQueryReturnManager()

    def get_state_key(self, event: CallbackQueryCute) -> int | None:
        return event.message_id.unwrap_or_none()


__all__ = ("CallbackQueryView",)



import typing
from functools import cached_property

import msgspec
from fntypes.result import Error, Ok, Result

from telegrinder.api.error import APIError
from telegrinder.api.response import APIResponse
from telegrinder.api.token import Token
from telegrinder.client import ABCClient, AiohttpClient
from telegrinder.model import DataConverter, decoder
from telegrinder.types.methods import APIMethods


def compose_data(
    client: ABCClient,
    data: dict[str, typing.Any],
    files: dict[str, tuple[str, bytes]],
) -> typing.Any:
    converter = DataConverter(_files=files.copy())
    return client.get_form(
        data={k: converter(v) for k, v in data.items()},
        files=converter.files,
    )


class API(APIMethods):
    """Bot API with available API methods and http client."""

    API_URL = "https://api.telegram.org/"
    API_FILE_URL = "https://api.telegram.org/file/"

    def __init__(self, token: Token, *, http: ABCClient | None = None) -> None:
        self.token = token
        self.http = http or AiohttpClient()
        super().__init__(self)

    def __repr__(self) -> str:
        return "<{}: token={!r}, http={!r}>".format(
            self.__class__.__name__,
            self.token,
            self.http,
        )

    @cached_property
    def id(self) -> int:
        return self.token.bot_id

    @property
    def request_url(self) -> str:
        return self.API_URL + f"bot{self.token}/"

    @property
    def request_file_url(self) -> str:
        return self.API_FILE_URL + f"bot{self.token}/"

    async def download_file(self, file_path: str) -> bytes:
        return await self.http.request_content(f"{self.request_file_url}/{file_path}")

    async def request(
        self,
        method: str,
        data: dict[str, typing.Any] | None = None,
        files: dict[str, tuple[str, bytes]] | None = None,
    ) -> Result[dict[str, typing.Any] | list[typing.Any] | bool, APIError]:
        response = await self.http.request_json(
            url=self.request_url + method,
            data=compose_data(self.http, data or {}, files or {}),
        )
        if response.get("ok"):
            assert "result" in response
            return Ok(response["result"])
        return Error(
            APIError(
                code=response.get("error_code", 400),
                error=response.get("description"),
            )
        )

    async def request_raw(
        self,
        method: str,
        data: dict[str, typing.Any] | None = None,
        files: dict[str, tuple[str, bytes]] | None = None,
    ) -> Result[msgspec.Raw, APIError]:
        response_bytes = await self.http.request_bytes(
            url=self.request_url + method,
            data=compose_data(self.http, data or {}, files or {}),
        )
        return decoder.decode(response_bytes, type=APIResponse).to_result()


__all__ = ("API",)



import abc
import enum
import typing
from dataclasses import dataclass

from fntypes import Option

from telegrinder.bot.rules.state import State, StateMeta

Payload = typing.TypeVar("Payload")


@dataclass(frozen=True, slots=True)
class StateData(typing.Generic[Payload]):
    key: str | enum.Enum
    payload: Payload


class ABCStateStorage(abc.ABC, typing.Generic[Payload]):
    @abc.abstractmethod
    async def get(self, user_id: int) -> Option[StateData[Payload]]: ...

    @abc.abstractmethod
    async def delete(self, user_id: int) -> None: ...

    @abc.abstractmethod
    async def set(self, user_id: int, key: str | enum.Enum, payload: Payload) -> None: ...

    def State(self, key: str | StateMeta | enum.Enum = StateMeta.ANY, /) -> State[Payload]:  # noqa: N802
        """Can be used as a shortcut to get a state rule dependant on current storage."""

        return State(storage=self, key=key)


__all__ = ("ABCStateStorage", "StateData")



from telegrinder.tools.state_storage.abc import ABCStateStorage, StateData
from telegrinder.tools.state_storage.memory import MemoryStateStorage

__all__ = ("ABCStateStorage", "MemoryStateStorage", "StateData")



from .buttons import BaseButton
from .error_handler import ABCErrorHandler, Catcher, CatcherError, ErrorHandler
from .formatting import (
    BaseSpecFormat,
    ChannelBoostLink,
    FormatString,
    HTMLFormatter,
    InviteChatLink,
    Link,
    Mention,
    PreCode,
    ResolveDomain,
    SpecialFormat,
    StartBotLink,
    StartGroupLink,
    TgEmoji,
    block_quote,
    bold,
    channel_boost_link,
    code_inline,
    escape,
    get_channel_boost_link,
    get_invite_chat_link,
    get_mention_link,
    get_resolve_domain_link,
    get_start_bot_link,
    get_start_group_link,
    invite_chat_link,
    italic,
    link,
    mention,
    pre_code,
    resolve_domain,
    spoiler,
    start_bot_link,
    start_group_link,
    strike,
    tg_emoji,
    underline,
)
from .global_context import (
    ABCGlobalContext,
    CtxVar,
    GlobalContext,
    GlobalCtxVar,
    TelegrinderContext,
    ctx_var,
)
from .i18n import (
    ABCI18n,
    ABCTranslator,
    ABCTranslatorMiddleware,
    I18nEnum,
    SimpleI18n,
    SimpleTranslator,
)
from .kb_set import KeyboardSetBase, KeyboardSetYAML
from .keyboard import (
    AnyMarkup,
    Button,
    InlineButton,
    InlineKeyboard,
    Keyboard,
    RowButtons,
)
from .limited_dict import LimitedDict
from .loop_wrapper import ABCLoopWrapper, DelayedTask, Lifespan, LoopWrapper
from .magic import impl, magic_bundle, resolve_arg_names
from .parse_mode import ParseMode
from .state_storage import ABCStateStorage, MemoryStateStorage, StateData

__all__ = (
    "ABCErrorHandler",
    "ABCGlobalContext",
    "ABCI18n",
    "ABCLoopWrapper",
    "ABCStateStorage",
    "ABCTranslator",
    "ABCTranslatorMiddleware",
    "AnyMarkup",
    "BaseButton",
    "BaseSpecFormat",
    "Button",
    "Catcher",
    "CatcherError",
    "ChannelBoostLink",
    "CtxVar",
    "DelayedTask",
    "ErrorHandler",
    "FormatString",
    "GlobalContext",
    "GlobalCtxVar",
    "HTMLFormatter",
    "I18nEnum",
    "InlineButton",
    "InlineKeyboard",
    "InviteChatLink",
    "Keyboard",
    "KeyboardSetBase",
    "KeyboardSetYAML",
    "Lifespan",
    "LimitedDict",
    "Link",
    "LoopWrapper",
    "MemoryStateStorage",
    "Mention",
    "ParseMode",
    "PreCode",
    "ResolveDomain",
    "RowButtons",
    "SimpleI18n",
    "SimpleTranslator",
    "SpecialFormat",
    "StartBotLink",
    "StartGroupLink",
    "StateData",
    "TelegrinderContext",
    "TgEmoji",
    "block_quote",
    "bold",
    "channel_boost_link",
    "code_inline",
    "ctx_var",
    "escape",
    "get_channel_boost_link",
    "get_invite_chat_link",
    "get_mention_link",
    "get_resolve_domain_link",
    "get_start_bot_link",
    "get_start_group_link",
    "impl",
    "invite_chat_link",
    "italic",
    "link",
    "magic_bundle",
    "mention",
    "pre_code",
    "resolve_arg_names",
    "resolve_domain",
    "spoiler",
    "start_bot_link",
    "start_group_link",
    "strike",
    "tg_emoji",
    "underline",
)



import enum


class ProgrammingLanguage(str, enum.Enum):
    """Type of ProgrammingLanguage."""

    ASSEMBLY = "assembly"
    PYTHON = "python"
    RUST = "rust"
    RUBY = "ruby"
    JAVA = "java"
    C = "c"
    C_SHARP = "cs"
    CPP = "cpp"
    CSS = "css"
    OBJECTIVE_C = "Objective-C"
    CYTHON = "cython"
    CARBON = "carbon"
    COBRA = "cobra"
    SWIFT = "swift"
    LAURELANG = "laurelang"
    DART = "dart"
    DELPHI = "delphi"
    ELIXIR = "elixir"
    HASKELL = "haskell"
    PASCAL = "pascal"
    TYPE_SCRIPT = "ts"
    JAVA_SCRIPT = "js"
    PHP = "php"
    GO = "go"
    SQL = "sql"
    F_SHARP = "fs"
    FORTRAN = "fortran"
    KOTLIN = "kotlin"
    HTML = "html"
    YAML = "yaml"
    JSON = "json"
    MARKDOWN = "markdown"
    MARKUP = "markup"


class ChatAction(str, enum.Enum):
    """Type of ChatAction.

    Choose one, depending on what the user is about to receive:
    - typing for text messages,
    - upload_photo for photos,
    - record_video or upload_video for videos,
    - record_voice or upload_voice for voice notes,
    - upload_document for general files,
    - choose_sticker for stickers,
    - find_location for location data,
    - record_video_note or upload_video_note for video notes.

    Docs: https://core.telegram.org/bots/api#sendchataction"""

    TYPING = "typing"
    UPLOAD_PHOTO = "upload_photo"
    RECORD_VIDEO = "record_video"
    UPLOAD_VIDEO = "upload_video"
    RECORD_VOICE = "record_voice"
    UPLOAD_VOICE = "upload_voice"
    UPLOAD_DOCUMENT = "upload_document"
    CHOOSE_STICKER = "choose_sticker"
    FIND_LOCATION = "find_location"
    RECORD_VIDEO_NOTE = "record_video_note"
    UPLOAD_VIDEO_NOTE = "upload_video_note"


class ReactionEmoji(str, enum.Enum):
    """Type of ReactionEmoji.

    Currently, it can be one of `👍`, `👎`, `❤`, `🔥`, `🥰`, `👏`,
    `😁`, `🤔`, `🤯`, `😱`, `🤬`, `😢`, `🎉`, `🤩`, `🤮`, `💩`, `🙏`, `👌`, `🕊`, `🤡`, `🥱`,
    `🥴`, `😍`, `🐳`, `❤‍🔥`, `🌚`, `🌭`, `💯`, `🤣`, `⚡`, `🍌`, `🏆`, `💔`, `🤨`, `😐`, `🍓`,
    `🍾`, `💋`, `🖕`, `😈`, `😴`, `😭`, `🤓`, `👻`, `👨‍💻`, `👀`, `🎃`, `🙈`, `😇`, `😨`, `🤝`,
    `✍`, `🤗`, `🫡`, `🎅`, `🎄`, `☃`, `💅`, `🤪`, `🗿`, `🆒`, `💘`, `🙉`, `🦄`, `😘`, `💊`,
    `🙊`, `😎`, `👾`, `🤷‍♂`, `🤷`, `🤷‍♀`, `😡`.

    Docs: https://core.telegram.org/bots/api#reactiontypeemoji"""

    THUMBS_UP = "👍"
    THUMBS_DOWN = "👎"
    RED_HEART = "❤"
    FIRE = "🔥"
    SMILING_FACE_WITH_HEARTS = "🥰"
    CLAPPING_HANDS = "👏"
    BEAMING_FACE_WITH_SMILING_EYES = "😁"
    THINKING_FACE = "🤔"
    EXPLODING_HEAD = "🤯"
    FACE_SCREAMING_IN_FEAR = "😱"
    FACE_WITH_SYMBOLS_ON_MOUTH = "🤬"
    CRYING_FACE = "😢"
    PARTY_POPPER = "🎉"
    STAR_STRUCK = "🤩"
    FACE_VOMITING = "🤮"
    PILE_OF_POO = "💩"
    FOLDED_HANDS = "🙏"
    OK_HAND = "👌"
    DOVE = "🕊"
    CLOWN_FACE = "🤡"
    YAWNING_FACE = "🥱"
    WOOZY_FACE = "🥴"
    SMILING_FACE_WITH_HEART_EYES = "😍"
    SPOUTING_WHALE = "🐳"
    HEART_ON_FIRE = "❤‍🔥"
    NEW_MOON_FACE = "🌚"
    HOT_DOG = "🌭"
    HUNDRED_POINTS = "💯"
    ROLLING_ON_THE_FLOOR_LAUGHING = "🤣"
    HIGH_VOLTAGE = "⚡"
    BANANA = "🍌"
    TROPHY = "🏆"
    BROKEN_HEART = "💔"
    FACE_WITH_RAISED_EYEBROW = "🤨"
    NEUTRAL_FACE = "😐"
    STRAWBERRY = "🍓"
    BOTTLE_WITH_POPPING_CORK = "🍾"
    KISS_MARK = "💋"
    MIDDLE_FINGER = "🖕"
    SMILING_FACE_WITH_HORNS = "😈"
    SLEEPING_FACE = "😴"
    LOUDLY_CRYING_FACE = "😭"
    NERD_FACE = "🤓"
    GHOST = "👻"
    MAN_TECHNOLOGIST = "👨‍💻"
    EYES = "👀"
    JACK_O_LANTERN = "🎃"
    SEE_NO_EVIL_MONKEY = "🙈"
    SMILING_FACE_WITH_HALO = "😇"
    FEARFUL_FACE = "😨"
    HANDSHAKE = "🤝"
    WRITING_HAND = "✍"
    SMILING_FACE_WITH_OPEN_HANDS = "🤗"
    SALUTING_FACE = "🫡"
    SANTA_CLAUS = "🎅"
    CHRISTMAS_TREE = "🎄"
    SNOWMAN = "☃"
    NAIL_POLISH = "💅"
    ZANY_FACE = "🤪"
    MOAI = "🗿"
    COOL_BUTTON = "🆒"
    HEART_WITH_ARROW = "💘"
    HEAR_NO_EVIL_MONKEY = "🙉"
    UNICORN = "🦄"
    FACE_BLOWING_A_KISS = "😘"
    PILL = "💊"
    SPEAK_NO_EVIL_MONKEY = "🙊"
    SMILING_FACE_WITH_SUNGLASSES = "😎"
    ALIEN_MONSTER = "👾"
    MAN_SHRUGGING = "🤷‍♂"
    PERSON_SHRUGGING = "🤷"
    WOMAN_SHRUGGING = "🤷‍♀"
    ENRAGED_FACE = "😡"


class DefaultAccentColor(int, enum.Enum):
    """Type of DefaultAccentColor.

    One of 7 possible user colors:
    - Red
    - Orange
    - Purple
    - Green
    - Cyan
    - Blue
    - Pink

    Docs: https://core.telegram.org/bots/api#accent-colors
    """

    RED = 0
    ORANGE = 1
    PURPLE = 2
    GREEN = 3
    CYAN = 4
    BLUE = 5
    PINK = 6


class TopicIconColor(int, enum.Enum):
    """Type of TopicIconColor.

    Docs: https://github.com/telegramdesktop/tdesktop/blob/991fe491c5ae62705d77aa8fdd44a79caf639c45/Telegram/SourceFiles/data/data_forum_topic.cpp#L51-L56
    """

    BLUE = 0x6FB9F0
    YELLOW = 0xFFD67E
    VIOLET = 0xCB86DB
    GREEN = 0x8EEE98
    ROSE = 0xFF93B2
    RED = 0xFB6F5F


class ChatBoostSourceType(str, enum.Enum):
    """Type of ChatBoostSourceType
    Docs: https://core.telegram.org/bots/api#chatboostsource"""

    PREMIUM = "premium"
    GIFT_CODE = "gift_code"
    GIVEAWAY = "giveaway"


class ContentType(str, enum.Enum):
    """Type of ContentType."""

    TEXT = "text"
    ANIMATION = "animation"
    AUDIO = "audio"
    DOCUMENT = "document"
    PHOTO = "photo"
    STICKER = "sticker"
    STORY = "story"
    VIDEO = "video"
    VIDEO_NOTE = "video_note"
    VOICE = "voice"
    HAS_MEDIA_SPOILER = "has_media_spoiler"
    CONTACT = "contact"
    DICE = "dice"
    GAME = "game"
    POLL = "poll"
    VENUE = "venue"
    LOCATION = "location"
    NEW_CHAT_MEMBERS = "new_chat_members"
    LEFT_CHAT_MEMBER = "left_chat_member"
    NEW_CHAT_TITLE = "new_chat_title"
    NEW_CHAT_PHOTO = "new_chat_photo"
    DELETE_CHAT_PHOTO = "delete_chat_photo"
    GROUP_CHAT_CREATED = "group_chat_created"
    BOOST_ADDED = "boost_added"
    SUPERGROUP_CHAT_CREATED = "supergroup_chat_created"
    CHANNEL_CHAT_CREATED = "channel_chat_created"
    MESSAGE_AUTO_DELETE_TIMER_CHANGED = "message_auto_delete_timer_changed"
    MIGRATE_TO_CHAT_ID = "migrate_to_chat_id"
    MIGRATE_FROM_CHAT_ID = "migrate_from_chat_id"
    PINNED_MESSAGE = "pinned_message"
    INVOICE = "invoice"
    SUCCESSFUL_PAYMENT = "successful_payment"
    USERS_SHARED = "users_shared"
    CHAT_SHARED = "chat_shared"
    CONNECTED_WEBSITE = "connected_website"
    WRITE_ACCESS_ALLOWED = "write_access_allowed"
    PASSPORT_DATA = "passport_data"
    PROXIMITY_ALERT_TRIGGERED = "proximity_alert_triggered"
    FORUM_TOPIC_CREATED = "forum_topic_created"
    FORUM_TOPIC_EDITED = "forum_topic_edited"
    FORUM_TOPIC_CLOSED = "forum_topic_closed"
    FORUM_TOPIC_REOPENED = "forum_topic_reopened"
    GENERAL_FORUM_TOPIC_HIDDEN = "general_forum_topic_hidden"
    GENERAL_FORUM_TOPIC_UNHIDDEN = "general_forum_topic_unhidden"
    GIVEAWAY_CREATED = "giveaway_created"
    GIVEAWAY = "giveaway"
    GIVEAWAY_WINNERS = "giveaway_winners"
    GIVEAWAY_COMPLETED = "giveaway_completed"
    VIDEO_CHAT_SCHEDULED = "video_chat_scheduled"
    VIDEO_CHAT_STARTED = "video_chat_started"
    VIDEO_CHAT_ENDED = "video_chat_ended"
    VIDEO_CHAT_PARTICIPANTS_INVITED = "video_chat_participants_invited"
    WEB_APP_DATA = "web_app_data"
    USER_SHARED = "user_shared"
    UNKNOWN = "unknown"


class Currency(str, enum.Enum):
    """Type of Currency.
    Docs: https://core.telegram.org/bots/payments#supported-currencies"""

    AED = "AED"
    AFN = "AFN"
    ALL = "ALL"
    AMD = "AMD"
    ARS = "ARS"
    AUD = "AUD"
    AZN = "AZN"
    BAM = "BAM"
    BDT = "BDT"
    BGN = "BGN"
    BND = "BND"
    BOB = "BOB"
    BRL = "BRL"
    BYN = "BYN"
    CAD = "CAD"
    CHF = "CHF"
    CLP = "CLP"
    CNY = "CNY"
    COP = "COP"
    CRC = "CRC"
    CZK = "CZK"
    DKK = "DKK"
    DOP = "DOP"
    DZD = "DZD"
    EGP = "EGP"
    ETB = "ETB"
    EUR = "EUR"
    GBP = "GBP"
    GEL = "GEL"
    GTQ = "GTQ"
    HKD = "HKD"
    HNL = "HNL"
    HRK = "HRK"
    HUF = "HUF"
    IDR = "IDR"
    ILS = "ILS"
    INR = "INR"
    ISK = "ISK"
    JMD = "JMD"
    JPY = "JPY"
    KES = "KES"
    KGS = "KGS"
    KRW = "KRW"
    KZT = "KZT"
    LBP = "LBP"
    LKR = "LKR"
    MAD = "MAD"
    MDL = "MDL"
    MNT = "MNT"
    MUR = "MUR"
    MVR = "MVR"
    MXN = "MXN"
    MYR = "MYR"
    MZN = "MZN"
    NGN = "NGN"
    NIO = "NIO"
    NOK = "NOK"
    NPR = "NPR"
    NZD = "NZD"
    PAB = "PAB"
    PEN = "PEN"
    PHP = "PHP"
    PKR = "PKR"
    PLN = "PLN"
    PYG = "PYG"
    QAR = "QAR"
    RON = "RON"
    RSD = "RSD"
    RUB = "RUB"
    SAR = "SAR"
    SEK = "SEK"
    SGD = "SGD"
    THB = "THB"
    TJS = "TJS"
    TRY = "TRY"
    TTD = "TTD"
    TWD = "TWD"
    TZS = "TZS"
    UAH = "UAH"
    UGX = "UGX"
    USD = "USD"
    UYU = "UYU"
    UZS = "UZS"
    VND = "VND"
    YER = "YER"
    ZAR = "ZAR"
    XTR = "XTR"
    """Telegram stars."""


class InlineQueryResultType(str, enum.Enum):
    """Type of InlineQueryResultType.
    Docs: https://core.telegram.org/bots/api#inlinequeryresult"""

    AUDIO = "audio"
    DOCUMENT = "document"
    GIF = "gif"
    MPEG4_GIF = "mpeg4_gif"
    PHOTO = "photo"
    STICKER = "sticker"
    VIDEO = "video"
    VOICE = "voice"
    ARTICLE = "article"
    CONTACT = "contact"
    GAME = "game"
    LOCATION = "location"
    VENUE = "venue"


class MenuButtonType(str, enum.Enum):
    """TType of MenuButtonType.
    Docs: https://core.telegram.org/bots/api#menubuttondefault
    """

    DEFAULT = "default"
    COMMANDS = "commands"
    WEB_APP = "web_app"


class InputMediaType(str, enum.Enum):
    """Type of InputMediaType.
    Docs: https://core.telegram.org/bots/api#inputmedia
    """

    ANIMATION = "animation"
    AUDIO = "audio"
    DOCUMENT = "document"
    PHOTO = "photo"
    VIDEO = "video"


class UpdateType(str, enum.Enum):
    """Type of update."""

    MESSAGE = "message"
    EDITED_MESSAGE = "edited_message"
    CHANNEL_POST = "channel_post"
    EDITED_CHANNEL_POST = "edited_channel_post"
    MESSAGE_REACTION = "message_reaction"
    MESSAGE_REACTION_COUNT = "message_reaction_count"
    INLINE_QUERY = "inline_query"
    CHOSEN_INLINE_RESULT = "chosen_inline_result"
    CALLBACK_QUERY = "callback_query"
    SHIPPING_QUERY = "shipping_query"
    PRE_CHECKOUT_QUERY = "pre_checkout_query"
    POLL = "poll"
    POLL_ANSWER = "poll_answer"
    MY_CHAT_MEMBER = "my_chat_member"
    CHAT_MEMBER = "chat_member"
    CHAT_JOIN_REQUEST = "chat_join_request"
    CHAT_BOOST = "chat_boost"
    REMOVED_CHAT_BOOST = "removed_chat_boost"
    BUSINESS_CONNECTION = "business_connection"
    BUSINESS_MESSAGE = "business_message"
    EDITED_BUSINESS_MESSAGE = "edited_business_message"
    DELETE_BUSINESS_MESSAGE = "delete_business_messages"


class BotCommandScopeType(str, enum.Enum):
    """Type of BotCommandScope.
    Represents the scope to which bot commands are applied."""

    DEFAULT = "default"
    ALL_PRIVATE_CHATS = "all_private_chats"
    ALL_GROUP_CHATS = "all_group_chats"
    ALL_CHAT_ADMINISTRATORS = "all_chat_administrators"
    CHAT = "chat"
    CHAT_ADMINISTRATORS = "chat_administrators"
    CHAT_MEMBER = "chat_member"


class ChatType(str, enum.Enum):
    """Type of chat, can be either `private`, `group`, `supergroup` or `channel`."""

    PRIVATE = "private"
    GROUP = "group"
    SUPERGROUP = "supergroup"
    CHANNEL = "channel"
    SENDER = "sender"


class ChatMemberStatus(str, enum.Enum):
    """Type of ChatMemberStatus."""

    CREATOR = "creator"
    ADMINISTRATOR = "administrator"
    MEMBER = "member"
    RESTRICTED = "restricted"
    LEFT = "left"
    KICKED = "kicked"


class DiceEmoji(str, enum.Enum):
    """Emoji on which the dice throw animation is based."""

    DICE = "🎲"
    DART = "🎯"
    BASKETBALL = "🏀"
    FOOTBALL = "⚽"
    SLOT_MACHINE = "🎰"
    BOWLING = "🎳"


class MessageEntityType(str, enum.Enum):
    """Type of the entity.
    Currently, can be `mention` (`@username`), `hashtag`
    (`#hashtag`), `cashtag` (`$USD`), `bot_command` (`/start@jobs_bot`),
    `url` (`https://telegram.org`), `email` (`do-not-reply@telegram.org`),
    `phone_number` (`+1-212-555-0123`), `bold` (**bold text**), `italic`
    (*italic text*), `underline` (underlined text), `strikethrough` (strikethrough
    text), `spoiler` (spoiler message), `code` (monowidth string), `pre`
    (monowidth block), `text_link` (for clickable text URLs), `text_mention`
    (for users [without usernames](https://telegram.org/blog/edit#new-mentions)),
    `custom_emoji` (for inline custom emoji stickers), `blockquote` (blockquote)
    [docs](https://core.telegram.org/bots/api#messageentity)"""

    MENTION = "mention"
    HASHTAG = "hashtag"
    CASHTAG = "cashtag"
    BOT_COMMAND = "bot_command"
    URL = "url"
    EMAIL = "email"
    PHONE_NUMBER = "phone_number"
    BOLD = "bold"
    ITALIC = "italic"
    UNDERLINE = "underline"
    STRIKETHROUGH = "strikethrough"
    SPOILER = "spoiler"
    BLOCKQUOTE = "blockquote"
    CODE = "code"
    PRE = "pre"
    TEXT_LINK = "text_link"
    TEXT_MENTION = "text_mention"
    CUSTOM_EMOJI = "custom_emoji"
    EXPANDABLE_BLOCKQUOTE = "expandable_blockquote"


class PollType(str, enum.Enum):
    """Poll type, currently can be `regular` or `quiz`."""

    REGULAR = "regular"
    QUIZ = "quiz"


class StickerType(str, enum.Enum):
    """Type of the sticker, currently one of `regular`, `mask`, `custom_emoji`.
    The type of the sticker is independent from its format, which is determined
    by the fields `is_animated` and `is_video`."""

    REGULAR = "regular"
    MASK = "mask"
    CUSTOM_EMOJI = "custom_emoji"


class MessageOriginType(str, enum.Enum):
    """Type of MessageOriginType
    Docs: https://core.telegram.org/bots/api#messageorigin"""

    USER = "user"
    HIDDEN_USER = "hidden_user"
    CHAT = "chat"
    CHANNEL = "channel"


class StickerSetStickerType(str, enum.Enum):
    """Type of stickers in the set, currently one of `regular`, `mask`, `custom_emoji`."""

    REGULAR = "regular"
    MASK = "mask"
    CUSTOM_EMOJI = "custom_emoji"


class MaskPositionPoint(str, enum.Enum):
    """The part of the face relative to which the mask should be placed. One of `forehead`,
    `eyes`, `mouth`, or `chin`."""

    FOREHEAD = "forehead"
    EYES = "eyes"
    MOUTH = "mouth"
    CHIN = "chin"


class InlineQueryChatType(str, enum.Enum):
    """Type of the chat from which the inline query was sent. Can be
    either `sender` for a private chat with the inline query sender, `private`,
    `group`, `supergroup`, or `channel`. The chat type should be always known
    for requests sent from official clients and most third-party clients,
    unless the request was sent from a secret chat."""

    SENDER = "sender"
    PRIVATE = "private"
    GROUP = "group"
    SUPERGROUP = "supergroup"
    CHANNEL = "channel"


class InlineQueryResultMimeType(str, enum.Enum):
    """MIME type of the content of the video URL, `text/html` or `video/mp4`."""

    TEXT_HTML = "text/html"
    VIDEO_MP4 = "video/mp4"


class InlineQueryResultThumbnailMimeType(str, enum.Enum):
    """MIME type of the thumbnail, must be one of `image/jpeg`, `image/gif`,
    or `video/mp4`. Defaults to `image/jpeg`
    """

    IMAGE_JPEG = "image/jpeg"
    IMAGE_GIF = "image/gif"
    VIDEO_MP4 = "video/mp4"


class PassportElementErrorType(str, enum.Enum):
    """Type of PassportElementErrorType.
    Docs: https://core.telegram.org/bots/api#passportelementerror
    """

    DATA = "data"
    FRONT_SIDE = "front_side"
    REVERSE_SIDE = "reverse_side"
    SELFIE = "selfie"
    FILE = "file"
    FILES = "files"
    TRANSLATION_FILE = "translation_file"
    TRANSLATION_FILES = "translation_files"
    UNSPECIFIED = "unspecified"


class ReactionTypeType(str, enum.Enum):
    """Type of ReactionTypeType.
    Docs: https://core.telegram.org/bots/api#reactiontype
    """

    EMOJI = "emoji"
    CUSTOM_EMOJI = "custom_emoji"


class InlineQueryResultGifThumbnailMimeType(str, enum.Enum):
    """MIME type of the thumbnail, must be one of `image/jpeg`, `image/gif`,
    or `video/mp4`. Defaults to `image/jpeg`.
    """

    IMAGE_JPEG = "image/jpeg"
    IMAGE_GIF = "image/gif"
    VIDEO_MP4 = "video/mp4"


class InlineQueryResultMpeg4GifThumbnailMimeType(str, enum.Enum):
    """MIME type of the thumbnail, must be one of `image/jpeg`, `image/gif`,
    or `video/mp4`. Defaults to `image/jpeg`.
    """

    IMAGE_JPEG = "image/jpeg"
    IMAGE_GIF = "image/gif"
    VIDEO_MP4 = "video/mp4"


class InlineQueryResultVideoMimeType(str, enum.Enum):
    """MIME type of the content of the video URL, `text/html` or `video/mp4`."""

    TEXT_HTML = "text/html"
    VIDEO_MP4 = "video/mp4"


class InlineQueryResultDocumentMimeType(str, enum.Enum):
    """MIME type of the content of the file, either `application/pdf` or `application/zip`."""

    APPLICATION_PDF = "application/pdf"
    APPLICATION_ZIP = "application/zip"


class EncryptedPassportElementType(str, enum.Enum):
    """Element type. One of `personal_details`, `passport`, `driver_license`,
    `identity_card`, `internal_passport`, `address`, `utility_bill`,
    `bank_statement`, `rental_agreement`, `passport_registration`,
    `temporary_registration`, `phone_number`, `email`.
    """

    PERSONAL_DETAILS = "personal_details"
    PASSPORT = "passport"
    DRIVER_LICENSE = "driver_license"
    IDENTITY_CARD = "identity_card"
    INTERNAL_PASSPORT = "internal_passport"
    ADDRESS = "address"
    UTILITY_BILL = "utility_bill"
    BANK_STATEMENT = "bank_statement"
    RENTAL_AGREEMENT = "rental_agreement"
    PASSPORT_REGISTRATION = "passport_registration"
    TEMPORARY_REGISTRATION = "temporary_registration"
    PHONE_NUMBER = "phone_number"
    EMAIL = "email"


class StickerFormat(str, enum.Enum):
    """Format of the sticker."""

    STATIC = "static"
    ANIMATED = "animated"
    VIDEO = "video"


__all__ = (
    "BotCommandScopeType",
    "ChatAction",
    "ChatBoostSourceType",
    "ChatMemberStatus",
    "ChatType",
    "ContentType",
    "Currency",
    "DefaultAccentColor",
    "DiceEmoji",
    "EncryptedPassportElementType",
    "InlineQueryChatType",
    "InlineQueryResultDocumentMimeType",
    "InlineQueryResultGifThumbnailMimeType",
    "InlineQueryResultMimeType",
    "InlineQueryResultMpeg4GifThumbnailMimeType",
    "InlineQueryResultThumbnailMimeType",
    "InlineQueryResultVideoMimeType",
    "MaskPositionPoint",
    "MessageEntityType",
    "MessageOriginType",
    "PassportElementErrorType",
    "PollType",
    "ProgrammingLanguage",
    "ReactionEmoji",
    "ReactionTypeType",
    "StickerFormat",
    "StickerSetStickerType",
    "StickerType",
    "TopicIconColor",
    "UpdateType",
)



import typing

from telegrinder.bot.dispatch.context import Context
from telegrinder.node.base import Node

from .abc import ABCRule
from .adapter.node import NodeAdapter


class NodeRule(ABCRule[tuple[Node, ...]]):
    def __init__(self, *nodes: type[Node] | tuple[str, type[Node]]) -> None:
        bindings = [binding if isinstance(binding, tuple) else (None, binding) for binding in nodes]
        self.nodes = [binding[1] for binding in bindings]
        self.node_keys = [binding[0] for binding in bindings]

    @property
    def adapter(self) -> NodeAdapter:
        return NodeAdapter(*self.nodes)  # type: ignore

    async def check(self, resolved_nodes: tuple[Node, ...], ctx: Context) -> typing.Literal[True]:
        for i, node in enumerate(resolved_nodes):
            if key := self.node_keys[i]:
                ctx[key] = node
        return True


__all__ = ("NodeRule",)



from .bot.rules import *  # noqa: F403

__all__ = (
    "ABCRule",
    "AndRule",
    "Argument",
    "CallbackDataEq",
    "CallbackDataJsonEq",
    "CallbackDataJsonModel",
    "CallbackDataMap",
    "CallbackDataMarkup",
    "CallbackQueryDataRule",
    "CallbackQueryRule",
    "ChatJoinRequestRule",
    "Command",
    "EnumTextRule",
    "FuncRule",
    "FuzzyText",
    "HasData",
    "HasEntities",
    "HasInviteLink",
    "HasLocation",
    "HasMention",
    "HasText",
    "InlineQueryChatType",
    "InlineQueryMarkup",
    "InlineQueryRule",
    "InlineQueryText",
    "IntegerInRange",
    "InviteLinkByCreator",
    "InviteLinkName",
    "IsBot",
    "IsChat",
    "IsChatId",
    "IsDice",
    "IsDiceEmoji",
    "IsForum",
    "IsForward",
    "IsForwardType",
    "IsGroup",
    "IsInteger",
    "IsLanguageCode",
    "IsPremium",
    "IsPrivate",
    "IsReply",
    "IsSuperGroup",
    "IsUpdateType",
    "IsUser",
    "IsUserId",
    "Markup",
    "MessageEntities",
    "MessageRule",
    "NodeRule",
    "NotRule",
    "OrRule",
    "Regex",
    "RuleEnum",
    "StartCommand",
    "Text",
)



import typing

from telegrinder.model import get_params
from telegrinder.types.objects import (
    InputFile,
    InputMediaAnimation,
    InputMediaAudio,
    InputMediaDocument,
    InputMediaPhoto,
    InputMediaVideo,
    LinkPreviewOptions,
    MessageEntity,
    ReactionEmoji,
    ReactionType,
    ReactionTypeEmoji,
    ReplyParameters,
)

InputMedia: typing.TypeAlias = typing.Union[
    InputMediaAnimation,
    InputMediaAudio,
    InputMediaDocument,
    InputMediaPhoto,
    InputMediaVideo,
]

INPUT_MEDIA_TYPES: typing.Final[dict[str, type[InputMedia]]] = {
    "animation": InputMediaAnimation,
    "audio": InputMediaAudio,
    "document": InputMediaDocument,
    "photo": InputMediaPhoto,
    "video": InputMediaVideo,
}


def compose_reactions(
    reactions: str | ReactionEmoji | ReactionType | list[str | ReactionEmoji | ReactionType],
    /,
) -> list[ReactionType]:
    if not isinstance(reactions, list):
        reactions = [reactions]
    return [
        (
            ReactionTypeEmoji("emoji", emoji)
            if isinstance(emoji, ReactionEmoji)
            else (ReactionTypeEmoji("emoji", ReactionEmoji(emoji)) if isinstance(emoji, str) else emoji)
        )
        for emoji in ([reactions] if isinstance(reactions, str) else reactions)  # type: ignore
    ]


def compose_reply_params(
    message_id: int | None,
    chat_id: int | str | None,
    *,
    allow_sending_without_reply: bool | None = None,
    quote: str | None = None,
    quote_parse_mode: str | None = None,
    quote_entities: list[MessageEntity] | None = None,
    quote_position: int | None = None,
    **other: typing.Any,
) -> ReplyParameters:
    return ReplyParameters(**get_params(locals()))


def compose_link_preview_options(
    *,
    is_disabled: bool | None = None,
    url: str | None = None,
    prefer_small_media: bool | None = None,
    prefer_large_media: bool | None = None,
    show_above_text: bool | None = None,
    **other: typing.Any,
) -> LinkPreviewOptions:
    return LinkPreviewOptions(**get_params(locals()))


def input_media(
    type: typing.Literal["animation", "audio", "document", "photo", "video"],
    media: str | InputFile,
    *,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    **other: typing.Any,
) -> InputMedia:
    return INPUT_MEDIA_TYPES[type](**get_params(locals()))


__all__ = (
    "compose_link_preview_options",
    "compose_reactions",
    "compose_reply_params",
    "input_media",
)



from telegrinder.api.api import API
from telegrinder.node.base import ComposeError, ScalarNode
from telegrinder.node.scope import GLOBAL
from telegrinder.types.objects import User


class Me(ScalarNode, User):
    scope = GLOBAL

    @classmethod
    async def compose(cls, api: API) -> User:
        me = await api.get_me()
        return me.expect(ComposeError("Can't complete get_me request"))


__all__ = ("Me",)



from telegrinder.node.base import ComposeError, ScalarNode
from telegrinder.node.message import MessageNode


class Text(ScalarNode, str):
    @classmethod
    async def compose(cls, message: MessageNode) -> str:
        if not message.text:
            raise ComposeError("Message has no text.")
        return message.text.unwrap()


class TextInteger(ScalarNode, int):
    @classmethod
    async def compose(cls, text: Text) -> int:
        if not text.isdigit():
            raise ComposeError("Text is not digit.")
        return int(text)


__all__ = ("Text", "TextInteger")



import typing

from telegrinder.api.api import API
from telegrinder.bot.cute_types.message import InputMediaType, MessageCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.base import BaseReplyHandler
from telegrinder.bot.rules.abc import ABCRule


class MediaGroupReplyHandler(BaseReplyHandler):
    def __init__(
        self,
        media: InputMediaType | list[InputMediaType],
        *rules: ABCRule,
        caption: str | list[str] | None = None,
        parse_mode: str | list[str] | None = None,
        is_blocking: bool = True,
        as_reply: bool = False,
        preset_context: Context | None = None,
        **default_params: typing.Any,
    ) -> None:
        self.media = media
        self.parse_mode = parse_mode
        self.caption = caption
        super().__init__(
            *rules,
            is_blocking=is_blocking,
            as_reply=as_reply,
            preset_context=preset_context,
            **default_params,
        )

    async def run(self, _: API, event: MessageCute, __: Context) -> typing.Any:
        method = event.answer_media_group if not self.as_reply else event.reply_media_group
        await method(
            media=self.media,
            parse_mode=self.parse_mode,
            caption=self.caption,
            **self.default_params,
        )


__all__ = ("MediaGroupReplyHandler",)



import enum
import typing
from reprlib import recursive_repr

from telegrinder.types.objects import Update

T = typing.TypeVar("T")

Key: typing.TypeAlias = str | enum.Enum
AnyValue: typing.TypeAlias = typing.Any


class Context(dict[str, AnyValue]):
    """Context class like dict & dotdict.

    For example:
    ```python
    class MyRule(ABCRule[T]):
        async def check(self, event: T, ctx: Context) -> bool:
            ctx.me = (await event.ctx_api.get_me()).unwrap()
            ctx["items"] = [1, 2, 3]
            return True
    ```
    """

    raw_update: Update

    def __init__(self, **kwargs: AnyValue) -> None:
        cls_vars = vars(self.__class__)
        defaults = {}

        for k in self.__class__.__annotations__:
            if k in cls_vars:
                defaults[k] = cls_vars[k]
                delattr(self.__class__, k)

        dict.__init__(self, **defaults | kwargs)

    @recursive_repr()
    def __repr__(self) -> str:
        return "{}({})".format(self.__class__.__name__, ", ".join(f"{k}={v!r}" for k, v in self.items()))

    def __setitem__(self, __key: Key, __value: AnyValue) -> None:
        dict.__setitem__(self, self.key_to_str(__key), __value)

    def __getitem__(self, __key: Key) -> AnyValue:
        return dict.__getitem__(self, self.key_to_str(__key))

    def __delitem__(self, __key: Key) -> None:
        dict.__delitem__(self, self.key_to_str(__key))

    def __setattr__(self, __name: str, __value: AnyValue) -> None:
        self.__setitem__(__name, __value)

    def __getattr__(self, __name: str) -> AnyValue:
        return self.__getitem__(__name)

    def __delattr__(self, __name: str) -> None:
        self.__delitem__(__name)

    @staticmethod
    def key_to_str(key: Key) -> str:
        return key if isinstance(key, str) else str(key.value)

    def copy(self) -> typing.Self:
        return self.__class__(**dict.copy(self))

    def set(self, key: Key, value: AnyValue) -> None:
        self[key] = value

    @typing.overload
    def get(self, key: Key) -> AnyValue | None: ...

    @typing.overload
    def get(self, key: Key, default: T) -> T | AnyValue: ...

    @typing.overload
    def get(self, key: Key, default: None = None) -> AnyValue | None: ...

    def get(self, key: Key, default: T | None = None) -> T | AnyValue | None:
        return dict.get(self, key, default)

    def get_or_set(self, key: Key, default: T) -> T:
        if key not in self:
            self.set(key, default)
        return self.get(key, default)

    def delete(self, key: Key) -> None:
        del self[key]


__all__ = ("Context",)



from telegrinder.bot.dispatch.abc import ABCDispatch
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.dispatch import Dispatch, TelegrinderContext
from telegrinder.bot.dispatch.handler import (
    ABCHandler,
    AudioReplyHandler,
    DocumentReplyHandler,
    FuncHandler,
    MediaGroupReplyHandler,
    MessageReplyHandler,
    PhotoReplyHandler,
    StickerReplyHandler,
    VideoReplyHandler,
)
from telegrinder.bot.dispatch.middleware import ABCMiddleware
from telegrinder.bot.dispatch.process import check_rule, process_inner
from telegrinder.bot.dispatch.return_manager import (
    ABCReturnManager,
    BaseReturnManager,
    CallbackQueryReturnManager,
    InlineQueryReturnManager,
    Manager,
    MessageReturnManager,
    register_manager,
)
from telegrinder.bot.dispatch.view import (
    ABCStateView,
    ABCView,
    BaseStateView,
    BaseView,
    CallbackQueryView,
    ChatJoinRequestView,
    ChatMemberView,
    InlineQueryView,
    MessageView,
    RawEventView,
    ViewBox,
)
from telegrinder.bot.dispatch.waiter_machine import (
    CALLBACK_QUERY_FOR_MESSAGE,
    CALLBACK_QUERY_FROM_CHAT,
    CALLBACK_QUERY_IN_CHAT_FOR_MESSAGE,
    MESSAGE_FROM_USER,
    MESSAGE_FROM_USER_IN_CHAT,
    MESSAGE_IN_CHAT,
    Hasher,
    ShortState,
    StateViewHasher,
    WaiterMachine,
    clear_wm_storage_worker,
)

__all__ = (
    "ABCDispatch",
    "ABCHandler",
    "ABCMiddleware",
    "ABCReturnManager",
    "ABCStateView",
    "ABCView",
    "AudioReplyHandler",
    "BaseReturnManager",
    "BaseStateView",
    "BaseView",
    "CALLBACK_QUERY_FOR_MESSAGE",
    "CALLBACK_QUERY_FROM_CHAT",
    "CALLBACK_QUERY_IN_CHAT_FOR_MESSAGE",
    "CallbackQueryReturnManager",
    "CallbackQueryView",
    "ChatJoinRequestView",
    "ChatMemberView",
    "Context",
    "Dispatch",
    "DocumentReplyHandler",
    "FuncHandler",
    "Hasher",
    "InlineQueryReturnManager",
    "InlineQueryView",
    "MESSAGE_FROM_USER",
    "MESSAGE_FROM_USER_IN_CHAT",
    "MESSAGE_IN_CHAT",
    "Manager",
    "MediaGroupReplyHandler",
    "MessageReplyHandler",
    "MessageReturnManager",
    "MessageView",
    "PhotoReplyHandler",
    "RawEventView",
    "ShortState",
    "StateViewHasher",
    "StickerReplyHandler",
    "TelegrinderContext",
    "VideoReplyHandler",
    "ViewBox",
    "WaiterMachine",
    "check_rule",
    "clear_wm_storage_worker",
    "clear_wm_storage_worker",
    "process_inner",
    "register_manager",
)



import typing
from abc import ABC, abstractmethod

from telegrinder.api.api import API
from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.dispatch.handler.abc import ABCHandler
from telegrinder.types.objects import Update

Event = typing.TypeVar("Event", bound=BaseCute)


class ABCView(ABC):
    def __repr__(self) -> str:
        return "<{}>".format(self.__class__.__name__)

    @abstractmethod
    async def check(self, event: Update) -> bool:
        pass

    @abstractmethod
    async def process(self, event: Update, api: API) -> bool:
        pass

    @abstractmethod
    def load(self, external: typing.Self) -> None:
        pass


class ABCEventRawView(ABCView, ABC, typing.Generic[Event]):
    handlers: list[ABCHandler[Event]]


class ABCStateView(ABCView, typing.Generic[Event]):
    @abstractmethod
    def get_state_key(self, event: Event) -> int | None:
        pass


__all__ = (
    "ABCEventRawView",
    "ABCStateView",
    "ABCView",
)



from telegrinder.bot.rules.adapter.abc import ABCAdapter, Event
from telegrinder.bot.rules.adapter.errors import AdapterError
from telegrinder.bot.rules.adapter.event import EventAdapter
from telegrinder.bot.rules.adapter.node import NodeAdapter
from telegrinder.bot.rules.adapter.raw_update import RawUpdateAdapter

__all__ = (
    "ABCAdapter",
    "AdapterError",
    "Event",
    "EventAdapter",
    "NodeAdapter",
    "RawUpdateAdapter",
)



import typing

from fntypes.result import Result

from telegrinder.api import API, APIError
from telegrinder.bot.cute_types.base import BaseCute, compose_method_params, shortcut
from telegrinder.model import get_params
from telegrinder.types import (
    InlineQuery,
    InlineQueryResult,
    InlineQueryResultsButton,
    User,
)


class InlineQueryCute(BaseCute[InlineQuery], InlineQuery, kw_only=True):
    api: API

    @property
    def from_user(self) -> User:
        return self.from_

    @shortcut("answer_inline_query", custom_params={"results"})
    async def answer(
        self,
        results: InlineQueryResult | list[InlineQueryResult],
        inline_query_id: str,
        cache_time: int | None = None,
        is_personal: bool | None = None,
        next_offset: str | None = None,
        button: InlineQueryResultsButton | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.answer_inline_query()`, see the [documentation](https://core.telegram.org/bots/api#answerinlinequery)

        Use this method to send answers to an inline query. On success, True is returned.
        No more than 50 results per query are allowed.
        :param inline_query_id: Unique identifier for the answered query.

        :param results: A JSON-serialized array of results for the inline query.

        :param cache_time: The maximum amount of time in seconds that the result of the inline querymay be cached on the server. Defaults to 300.

        :param is_personal: Pass True if results may be cached on the server side only for the user thatsent the query. By default, results may be returned to any user who sendsthe same query.

        :param next_offset: Pass the offset that a client should send in the next query with the same textto receive more results. Pass an empty string if there are no more resultsor if you don't support pagination. Offset length can't exceed 64 bytes.
        :param button: A JSON-serialized object describing a button to be shown above inline queryresults."""

        params = compose_method_params(
            get_params(locals()),
            self,
            default_params={("inline_query_id", "id")},
        )
        params["results"] = [results] if not isinstance(results, list) else results
        return await self.ctx_api.answer_inline_query(**params)


__all__ = ("InlineQueryCute",)



from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.node.base import ComposeError, ScalarNode
from telegrinder.node.update import UpdateNode


class MessageNode(ScalarNode, MessageCute):
    @classmethod
    async def compose(cls, update: UpdateNode) -> MessageCute:
        if not update.message:
            raise ComposeError("Update is not a message.")
        return update.message.unwrap()


__all__ = ("MessageNode",)



class APIError(BaseException):
    def __init__(self, code: int, error: str | None = None) -> None:
        self.code, self.error = code, error

    def __str__(self) -> str:
        return f"[{self.code}] {self.error or 'Something went wrong'}"

    def __repr__(self) -> str:
        return f"<APIError: {self.__str__()}>"


class InvalidTokenError(BaseException):
    pass


__all__ = ("APIError", "InvalidTokenError")



from .callback import CALLBACK_QUERY_FOR_MESSAGE, CALLBACK_QUERY_FROM_CHAT, CALLBACK_QUERY_IN_CHAT_FOR_MESSAGE
from .hasher import Hasher
from .message import MESSAGE_FROM_USER, MESSAGE_FROM_USER_IN_CHAT, MESSAGE_IN_CHAT
from .state import StateViewHasher

__all__ = (
    "CALLBACK_QUERY_FOR_MESSAGE",
    "CALLBACK_QUERY_FROM_CHAT",
    "CALLBACK_QUERY_IN_CHAT_FOR_MESSAGE",
    "Hasher",
    "MESSAGE_FROM_USER",
    "MESSAGE_FROM_USER_IN_CHAT",
    "MESSAGE_IN_CHAT",
    "StateViewHasher",
)



import typing
from functools import cached_property

from fntypes.co import Nothing, Some

from telegrinder.api.api import API
from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.cute_types.callback_query import CallbackQueryCute
from telegrinder.bot.cute_types.chat_join_request import ChatJoinRequestCute
from telegrinder.bot.cute_types.chat_member_updated import ChatMemberUpdatedCute
from telegrinder.bot.cute_types.inline_query import InlineQueryCute
from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.msgspec_utils import Option
from telegrinder.types.objects import Model, Update

EventModel = typing.TypeVar("EventModel", bound=Model)


class UpdateCute(BaseCute[Update], Update, kw_only=True):
    api: API

    message: Option[MessageCute] = Nothing()
    """Optional. New incoming message of any kind - text, photo, sticker, etc."""

    edited_message: Option[MessageCute] = Nothing()
    """Optional. New version of a message that is known to the bot and was edited.
    This update may at times be triggered by changes to message fields that are
    either unavailable or not actively used by your bot."""

    channel_post: Option[MessageCute] = Nothing()
    """Optional. New incoming channel post of any kind - text, photo, sticker,
    etc."""

    edited_channel_post: Option[MessageCute] = Nothing()
    """Optional. New version of a channel post that is known to the bot and was edited.
    This update may at times be triggered by changes to message fields that are
    either unavailable or not actively used by your bot."""

    business_message: Option[MessageCute] = Nothing()
    """Optional. New message from a connected business account."""

    edited_business_message: Option[MessageCute] = Nothing()
    """Optional. New version of a message from a connected business account."""

    inline_query: Option[InlineQueryCute] = Nothing()
    """Optional. New incoming inline query."""

    callback_query: Option[CallbackQueryCute] = Nothing()
    """Optional. New incoming callback query."""

    my_chat_member: Option[ChatMemberUpdatedCute] = Nothing()
    """Optional. The bot's chat member status was updated in a chat. For private
    chats, this update is received only when the bot is blocked or unblocked
    by the user."""

    chat_member: Option[ChatMemberUpdatedCute] = Nothing()
    """Optional. A chat member's status was updated in a chat. The bot must be an
    administrator in the chat and must explicitly specify `chat_member` in
    the list of allowed_updates to receive these updates."""

    chat_join_request: Option[ChatJoinRequestCute] = Nothing()
    """Optional. A request to join the chat has been sent. The bot must have the can_invite_users
    administrator right in the chat to receive these updates."""

    @cached_property
    def incoming_update(self) -> Model:
        return getattr(self, self.update_type.value).unwrap()

    def get_event(self, event_model: type[EventModel]) -> Option[EventModel]:
        if isinstance(self.incoming_update, event_model):
            return Some(self.incoming_update)
        return Nothing()


__all__ = ("UpdateCute",)



def get_mention_link(user_id: int) -> str:
    return f"tg://user?id={user_id}"


def get_resolve_domain_link(username: str) -> str:
    return f"tg://resolve?domain={username}"


def get_start_bot_link(bot_id: str | int, data: str) -> str:
    if isinstance(bot_id, int):
        return get_mention_link(bot_id) + f"&start={data}"
    return get_resolve_domain_link(bot_id) + f"&start={data}"


def get_start_group_link(bot_id: str | int, data: str) -> str:
    if isinstance(bot_id, int):
        return get_mention_link(bot_id) + f"&startgroup={data}"
    return get_resolve_domain_link(bot_id) + f"&startgroup={data}"


def get_channel_boost_link(channel_id: str | int) -> str:
    if isinstance(channel_id, int):
        return get_mention_link(channel_id) + "&boost"
    return get_resolve_domain_link(channel_id) + "&boost"


def get_invite_chat_link(invite_link: str) -> str:
    return f"tg://join?invite={invite_link}"


__all__ = (
    "get_channel_boost_link",
    "get_invite_chat_link",
    "get_mention_link",
    "get_resolve_domain_link",
    "get_start_bot_link",
    "get_start_group_link",
)



from telegrinder.bot.dispatch.view.abc import ABCStateView, ABCView
from telegrinder.bot.dispatch.view.base import BaseStateView, BaseView
from telegrinder.bot.dispatch.view.box import ViewBox
from telegrinder.bot.dispatch.view.callback_query import CallbackQueryView
from telegrinder.bot.dispatch.view.chat_join_request import ChatJoinRequestView
from telegrinder.bot.dispatch.view.chat_member import ChatMemberView
from telegrinder.bot.dispatch.view.inline_query import InlineQueryView
from telegrinder.bot.dispatch.view.message import MessageView
from telegrinder.bot.dispatch.view.raw import RawEventView

__all__ = (
    "ABCStateView",
    "ABCView",
    "BaseStateView",
    "BaseView",
    "CallbackQueryView",
    "ChatJoinRequestView",
    "ChatMemberView",
    "InlineQueryView",
    "MessageView",
    "RawEventView",
    "ViewBox",
)



from telegrinder import node
from telegrinder.tools.i18n.base import ABCTranslator

from .abc import ABCRule, with_caching_translations
from .node import NodeRule


class HasText(NodeRule):
    def __init__(self) -> None:
        super().__init__(node.text.Text)


class Text(ABCRule):
    def __init__(self, texts: str | list[str], *, ignore_case: bool = False) -> None:
        if not isinstance(texts, list):
            texts = [texts]
        self.texts = texts if not ignore_case else list(map(str.lower, texts))
        self.ignore_case = ignore_case

    async def check(self, text: node.text.Text) -> bool:
        return (text if not self.ignore_case else text.lower()) in self.texts

    @with_caching_translations
    async def translate(self, translator: ABCTranslator) -> "Text":
        return Text(
            texts=[translator.get(text) for text in self.texts],
            ignore_case=self.ignore_case,
        )


__all__ = ("HasText", "Text")



import dataclasses
import typing

from fntypes.option import Nothing, Option

from telegrinder.api.api import API
from telegrinder.bot.cute_types import ChatJoinRequestCute
from telegrinder.node.base import ComposeError, DataNode, ScalarNode
from telegrinder.node.callback_query import CallbackQueryNode
from telegrinder.node.event import EventNode
from telegrinder.node.message import MessageNode
from telegrinder.node.polymorphic import Polymorphic, impl
from telegrinder.types.objects import Chat, Message, User


@dataclasses.dataclass(kw_only=True, slots=True)
class Source(Polymorphic, DataNode):
    api: API
    chat: Chat
    from_user: User
    thread_id: Option[int] = dataclasses.field(default_factory=lambda: Nothing())

    @impl
    async def compose_message(cls, message: MessageNode) -> typing.Self:
        return cls(
            api=message.ctx_api,
            chat=message.chat,
            from_user=message.from_.expect(ComposeError("MessageNode has no from_user")),
            thread_id=message.message_thread_id,
        )

    @impl
    async def compose_callback_query(cls, callback_query: CallbackQueryNode) -> typing.Self:
        return cls(
            api=callback_query.ctx_api,
            chat=callback_query.chat.expect(ComposeError("CallbackQueryNode has no chat")),
            from_user=callback_query.from_user,
            thread_id=callback_query.message_thread_id,
        )

    @impl
    async def compose_chat_join_request(
        cls, chat_join_request: EventNode[ChatJoinRequestCute]
    ) -> typing.Self:
        return cls(
            api=chat_join_request.ctx_api,
            chat=chat_join_request.chat,
            from_user=chat_join_request.from_user,
            thread_id=Nothing(),
        )

    async def send(self, text: str) -> Message:
        result = await self.api.send_message(
            chat_id=self.chat.id,
            message_thread_id=self.thread_id.unwrap_or_none(),
            text=text,
        )
        return result.unwrap()


class ChatSource(ScalarNode, Chat):
    @classmethod
    async def compose(cls, source: Source) -> Chat:
        return source.chat


class UserSource(ScalarNode, User):
    @classmethod
    async def compose(cls, source: Source) -> User:
        return source.from_user


__all__ = ("ChatSource", "Source", "UserSource")



import abc

from telegrinder.bot.dispatch.context import Context
from telegrinder.types.objects import Message as MessageEvent

from .abc import ABCRule, Message
from .adapter import EventAdapter


class MessageRule(ABCRule[Message], abc.ABC):
    adapter: EventAdapter[Message] = EventAdapter(MessageEvent, Message)

    @abc.abstractmethod
    async def check(self, message: Message, ctx: Context) -> bool: ...


__all__ = ("MessageRule",)



import asyncio
import datetime
import typing

from telegrinder.bot.dispatch.abc import ABCDispatch
from telegrinder.bot.dispatch.view.base import BaseStateView, BaseView
from telegrinder.bot.dispatch.waiter_machine.middleware import WaiterMiddleware
from telegrinder.bot.dispatch.waiter_machine.short_state import (
    EventModel,
    ShortState,
    ShortStateContext,
)
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.tools.limited_dict import LimitedDict

from .actions import WaiterActions
from .hasher import Hasher, StateViewHasher

T = typing.TypeVar("T")
HasherData = typing.TypeVar("HasherData")


Storage: typing.TypeAlias = dict[
    Hasher[EventModel, HasherData], LimitedDict[typing.Hashable, ShortState[EventModel]]
]

WEEK: typing.Final[datetime.timedelta] = datetime.timedelta(days=7)


class WaiterMachine:
    def __init__(
        self,
        dispatch: ABCDispatch | None = None,
        *,
        max_storage_size: int = 1000,
        base_state_lifetime: datetime.timedelta = WEEK,
    ) -> None:
        self.dispatch = dispatch
        self.max_storage_size = max_storage_size
        self.base_state_lifetime = base_state_lifetime
        self.storage: Storage = {}

    def __repr__(self) -> str:
        return "<{}: max_storage_size={}, {}>".format(
            self.__class__.__name__,
            self.max_storage_size,
            ", ".join(
                f"{view_name}: {len(self.storage[view_name].values())} shortstates"
                for view_name in self.storage
            )
            or "empty",
        )

    async def drop(
        self,
        hasher: Hasher[EventModel, HasherData],
        id: HasherData,
        **context: typing.Any,
    ) -> None:
        if hasher not in self.storage:
            raise LookupError("No record of hasher {!r} found.".format(hasher))

        waiter_id: typing.Hashable = hasher.get_hash_from_data(id).expect(
            RuntimeError("Couldn't create hash from data")
        )
        short_state = self.storage[hasher].pop(waiter_id, None)
        if short_state is None:
            raise LookupError(
                "Waiter with identificator {} is not found for hasher {!r}".format(waiter_id, hasher)
            )

        if on_drop := short_state.actions.get("on_drop"):
            on_drop(short_state, **context)

        short_state.cancel()

    async def drop_all(self) -> None:
        """Drops all waiters in storage."""

        for hasher in self.storage:
            for ident, short_state in self.storage[hasher].items():
                if short_state.context:
                    await self.drop(
                        hasher,
                        ident,
                    )
                else:
                    short_state.cancel()

    async def wait_from_event(
        self,
        view: BaseStateView[EventModel],
        event: EventModel,
        *,
        filter: ABCRule | None = None,
        release: ABCRule | None = None,
        lifetime: datetime.timedelta | float | None = None,
        **actions: typing.Unpack[WaiterActions],
    ) -> ShortStateContext[EventModel]:
        hasher = StateViewHasher(view.__class__)
        return await self.wait(
            hasher=hasher,
            data=hasher.get_data_from_event(event).expect(
                RuntimeError("Hasher couldn't create data from event.")
            ),
            filter=filter,
            release=release,
            lifetime=lifetime,
            **actions,
        )

    async def wait(
        self,
        hasher: Hasher[EventModel, HasherData],
        data: HasherData,
        *,
        filter: ABCRule | None = None,
        release: ABCRule | None = None,
        lifetime: datetime.timedelta | float | None = None,
        **actions: typing.Unpack[WaiterActions],
    ) -> ShortStateContext[EventModel]:
        if isinstance(lifetime, int | float):
            lifetime = datetime.timedelta(seconds=lifetime)

        event = asyncio.Event()
        short_state = ShortState[EventModel](
            filter=filter,
            release=release,
            event=event,
            lifetime=lifetime or self.base_state_lifetime,
            actions=actions,
        )
        waiter_hash = hasher.get_hash_from_data(data).expect(RuntimeError("Hasher couldn't create hash."))

        if hasher not in self.storage:
            if self.dispatch:
                view: BaseView[EventModel] = self.dispatch.get_view(hasher.view).expect(
                    RuntimeError(f"View {hasher.view.__name__} is not defined in dispatch")
                )
                view.middlewares.insert(0, WaiterMiddleware(self, hasher))
            self.storage[hasher] = LimitedDict(maxlimit=self.max_storage_size)

        if (deleted_short_state := self.storage[hasher].set(waiter_hash, short_state)) is not None:
            deleted_short_state.cancel()

        await event.wait()
        self.storage[hasher].pop(waiter_hash, None)
        assert short_state.context is not None
        return short_state.context

    async def clear_storage(
        self,
    ) -> None:
        """Clears storage."""

        for hasher in self.storage:
            now = datetime.datetime.now()
            for ident, short_state in self.storage.get(hasher, {}).copy().items():
                if short_state.expiration_date is not None and now > short_state.expiration_date:
                    await self.drop(
                        hasher,
                        ident,
                        force=True,
                    )


async def clear_wm_storage_worker(
    wm: WaiterMachine,
    interval_seconds: int = 60,
) -> typing.NoReturn:
    while True:
        await wm.clear_storage()
        await asyncio.sleep(interval_seconds)


__all__ = ("WaiterMachine",)



import dataclasses
import enum
import keyword
import secrets
import typing
from datetime import datetime
from types import NoneType

import msgspec
from fntypes.co import Nothing, Result, Some

from telegrinder.msgspec_utils import decoder, encoder, get_origin

if typing.TYPE_CHECKING:
    from telegrinder.api.error import APIError

T = typing.TypeVar("T")

UnionType: typing.TypeAlias = typing.Annotated[tuple[T, ...], ...]

MODEL_CONFIG: typing.Final[dict[str, typing.Any]] = {
    "omit_defaults": True,
    "dict": True,
    "rename": {kw + "_": kw for kw in keyword.kwlist},
}


@typing.overload
def full_result(
    result: Result[msgspec.Raw, "APIError"],
    full_t: type[T],
) -> Result[T, "APIError"]: ...


@typing.overload
def full_result(
    result: Result[msgspec.Raw, "APIError"],
    full_t: UnionType[T],
) -> Result[T, "APIError"]: ...


def full_result(
    result: Result[msgspec.Raw, "APIError"],
    full_t: typing.Any,
) -> Result[typing.Any, "APIError"]:
    return result.map(lambda v: decoder.decode(v, type=full_t))


def get_params(params: dict[str, typing.Any]) -> dict[str, typing.Any]:
    validated_params = {}
    for k, v in (
        *params.pop("other", {}).items(),
        *params.items(),
    ):
        if isinstance(v, Proxy):
            v = v.get()
        if k == "self" or type(v) in (NoneType, Nothing):
            continue
        validated_params[k] = v.unwrap() if isinstance(v, Some) else v
    return validated_params


class Model(msgspec.Struct, **MODEL_CONFIG):
    @classmethod
    def from_data(cls, data: dict[str, typing.Any]) -> typing.Self:
        return decoder.convert(data, type=cls)

    @classmethod
    def from_bytes(cls, data: bytes) -> typing.Self:
        return decoder.decode(data, type=cls)

    def _to_dict(
        self,
        dct_name: str,
        exclude_fields: set[str],
        full: bool,
    ) -> dict[str, typing.Any]:
        if dct_name not in self.__dict__:
            self.__dict__[dct_name] = (
                msgspec.structs.asdict(self)
                if not full
                else encoder.to_builtins(self.to_dict(exclude_fields=exclude_fields), order="deterministic")
            )

        if not exclude_fields:
            return self.__dict__[dct_name]

        return {key: value for key, value in self.__dict__[dct_name].items() if key not in exclude_fields}

    def to_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
    ) -> dict[str, typing.Any]:
        """
        :param exclude_fields: Model field names to exclude from the dictionary representation of this model.
        :return: A dictionary representation of this model.
        """

        return self._to_dict("model_as_dict", exclude_fields or set(), full=False)

    def to_full_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
    ) -> dict[str, typing.Any]:
        """
        :param exclude_fields: Model field names to exclude from the dictionary representation of this model.
        :return: A dictionary representation of this model including all models, structs, custom types.
        """

        return self._to_dict("model_as_full_dict", exclude_fields or set(), full=True)


@dataclasses.dataclass(kw_only=True, frozen=True, slots=True, repr=False)
class DataConverter:
    _converters: dict[type[typing.Any], typing.Callable[..., typing.Any]] = dataclasses.field(
        init=False,
        default_factory=lambda: {},
    )
    _files: dict[str, tuple[str, bytes]] = dataclasses.field(default_factory=lambda: {})

    def __repr__(self) -> str:
        return "<{}: {}>".format(
            self.__class__.__name__,
            ", ".join(f"{k}={v.__name__!r}" for k, v in self._converters.items()),
        )

    def __post_init__(self) -> None:
        self._converters.update(
            {
                get_origin(value.__annotations__["data"]): value
                for key, value in vars(self.__class__).items()
                if key.startswith("convert_") and callable(value)
            }
        )

    def __call__(self, data: typing.Any, *, serialize: bool = True) -> typing.Any:
        converter = self.get_converter(get_origin(type(data)))
        if converter is not None:
            if isinstance(converter, staticmethod):
                return converter(data, serialize)
            return converter(self, data, serialize)
        return data

    @property
    def converters(self) -> dict[type[typing.Any], typing.Callable[..., typing.Any]]:
        return self._converters.copy()

    @property
    def files(self) -> dict[str, tuple[str, bytes]]:
        return self._files.copy()

    @staticmethod
    def convert_enum(data: enum.Enum, _: bool = False) -> typing.Any:
        return data.value

    @staticmethod
    def convert_datetime(data: datetime, _: bool = False) -> int:
        return int(data.timestamp())

    def get_converter(self, t: type[typing.Any]):
        for type_, converter in self._converters.items():
            if issubclass(t, type_):
                return converter
        return None

    def convert_model(
        self,
        data: Model,
        serialize: bool = True,
    ) -> str | dict[str, typing.Any]:
        converted_dct = self(data.to_dict(), serialize=False)
        return encoder.encode(converted_dct) if serialize is True else converted_dct

    def convert_dct(
        self,
        data: dict[str, typing.Any],
        serialize: bool = True,
    ) -> dict[str, typing.Any]:
        return {
            k: self(v, serialize=serialize) for k, v in data.items() if type(v) not in (NoneType, Nothing)
        }

    def convert_lst(
        self,
        data: list[typing.Any],
        serialize: bool = True,
    ) -> str | list[typing.Any]:
        converted_lst = [self(x, serialize=False) for x in data]
        return encoder.encode(converted_lst) if serialize is True else converted_lst

    def convert_tpl(self, data: tuple[typing.Any, ...], _: bool = False) -> str | tuple[typing.Any, ...]:
        match data:
            case (str(filename), bytes(content)):
                attach_name = secrets.token_urlsafe(16)
                self._files[attach_name] = (filename, content)
                return "attach://{}".format(attach_name)

        return data


class Proxy:
    def __init__(self, cfg: "_ProxiedDict", key: str) -> None:
        self.key = key
        self.cfg = cfg

    def get(self) -> typing.Any | None:
        return self.cfg._defaults.get(self.key)


class _ProxiedDict(typing.Generic[T]):
    def __init__(self, tp: type[T]) -> None:
        self.type = tp
        self._defaults = {}

    def __setattribute__(self, name: str, value: typing.Any, /) -> None:
        self._defaults[name] = value

    def __getitem__(self, key: str, /) -> None:
        return Proxy(self, key)  # type: ignore

    def __setitem__(self, key: str, value: typing.Any, /) -> None:
        self._defaults[key] = value


if typing.TYPE_CHECKING:

    def ProxiedDict(typed_dct: type[T]) -> T | _ProxiedDict[T]:  # noqa: N802
        ...

else:
    ProxiedDict = _ProxiedDict


__all__ = (
    "DataConverter",
    "MODEL_CONFIG",
    "Model",
    "ProxiedDict",
    "Proxy",
    "full_result",
    "get_params",
)



import dataclasses
from functools import cached_property

import typing_extensions as typing

from telegrinder.api.api import API
from telegrinder.bot.cute_types import BaseCute, UpdateCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.process import check_rule
from telegrinder.model import Model
from telegrinder.modules import logger
from telegrinder.node.base import Node, get_nodes
from telegrinder.node.composer import compose_nodes
from telegrinder.node.event import EVENT_NODE_KEY
from telegrinder.tools.error_handler import ABCErrorHandler, ErrorHandler
from telegrinder.types.enums import UpdateType
from telegrinder.types.objects import Update

from .abc import ABCHandler

if typing.TYPE_CHECKING:
    from telegrinder.bot.rules import ABCRule

F = typing.TypeVar(
    "F",
    bound=typing.Callable[typing.Concatenate[typing.Any, ...], typing.Awaitable[typing.Any]],
)
Event = typing.TypeVar("Event", bound=Model)
ErrorHandlerT = typing.TypeVar("ErrorHandlerT", bound=ABCErrorHandler, default=ErrorHandler)


@dataclasses.dataclass(repr=False, slots=True)
class FuncHandler(ABCHandler[Event], typing.Generic[Event, F, ErrorHandlerT]):
    func: F
    rules: list["ABCRule"]
    is_blocking: bool = dataclasses.field(default=True, kw_only=True)
    dataclass: type[typing.Any] | None = dataclasses.field(default=dict, kw_only=True)
    error_handler: ErrorHandlerT = dataclasses.field(
        default_factory=lambda: typing.cast(ErrorHandlerT, ErrorHandler()),
        kw_only=True,
    )
    preset_context: Context = dataclasses.field(default_factory=lambda: Context(), kw_only=True)
    update_type: UpdateType | None = dataclasses.field(default=None, kw_only=True)

    def __repr__(self) -> str:
        return "<{}: {}={!r} with rules={!r}, dataclass={!r}, error_handler={!r}>".format(
            self.__class__.__name__,
            "blocking function" if self.is_blocking else "function",
            self.func.__name__,
            self.rules,
            self.dataclass,
            self.error_handler,
        )

    @cached_property
    def required_nodes(self) -> dict[str, type[Node]]:
        return get_nodes(self.func)

    async def check(self, api: API, event: Update, ctx: Context | None = None) -> bool:
        if self.update_type is not None and self.update_type != event.update_type:
            return False

        logger.debug("Checking handler {!r}...", self)
        ctx = Context(raw_update=event) if ctx is None else ctx
        temp_ctx = ctx.copy()
        temp_ctx |= self.preset_context

        nodes = self.required_nodes
        node_col = None
        update = event

        if nodes:
            result = await compose_nodes(nodes, ctx, data={Update: event, API: api})
            if not result:
                logger.debug(f"Cannot compose nodes for handler. Error: {result.error!r}")
                return False

            node_col = result.value
            temp_ctx |= node_col.values

            if EVENT_NODE_KEY in ctx:
                for name, node in nodes.items():
                    if node is ctx[EVENT_NODE_KEY] and name in temp_ctx:
                        ctx[EVENT_NODE_KEY] = temp_ctx.pop(name)

        for rule in self.rules:
            if not await check_rule(api, rule, update, temp_ctx):
                logger.debug("Rule {!r} failed!", rule)
                return False

        logger.debug("All checks passed for handler.")

        temp_ctx["node_col"] = node_col
        ctx |= temp_ctx
        return True

    async def run(self, api: API, event: Event, ctx: Context) -> typing.Any:
        logger.debug(f"Running func handler {self.func.__qualname__!r}")
        dataclass_type = typing.get_origin(self.dataclass) or self.dataclass

        if dataclass_type is Update and (event_node := ctx.pop(EVENT_NODE_KEY, None)) is not None:
            event = event_node

        elif dataclass_type is not None:
            if self.update_type is not None and isinstance(event, Update):
                update = event.to_dict()[self.update_type.value].unwrap()
                event = (
                    self.dataclass.from_update(update, bound_api=api)  # type: ignore
                    if issubclass(dataclass_type, BaseCute)
                    else self.dataclass(**update.to_dict())  # type: ignore
                )

            elif issubclass(dataclass_type, UpdateCute) and isinstance(event, Update):
                event = self.dataclass.from_update(event, bound_api=api)  # type: ignore

            else:
                event = self.dataclass(**event.to_dict())  # type: ignore

        result = (await self.error_handler.run(self.func, event, api, ctx)).unwrap()
        if node_col := ctx.node_col:
            await node_col.close_all()
        return result


__all__ = ("FuncHandler",)



import typing
from contextlib import suppress

import msgspec
from fntypes.co import Nothing, Result, Some, Variative, unwrapping

from telegrinder.api import API, APIError
from telegrinder.bot.cute_types.base import BaseCute, compose_method_params, shortcut
from telegrinder.bot.cute_types.message import MediaType, MessageCute, ReplyMarkup, execute_method_edit
from telegrinder.model import get_params
from telegrinder.msgspec_utils import Option, decoder
from telegrinder.types.objects import (
    CallbackQuery,
    Chat,
    InaccessibleMessage,
    InlineKeyboardMarkup,
    InputFile,
    InputMedia,
    LinkPreviewOptions,
    MessageEntity,
    MessageId,
    ReplyParameters,
    User,
)


class CallbackQueryCute(BaseCute[CallbackQuery], CallbackQuery, kw_only=True):
    api: API

    message: Option[Variative[MessageCute, InaccessibleMessage]] = Nothing()
    """Optional. Message sent by the bot with the callback button that originated
    the query."""

    @property
    def from_user(self) -> User:
        return self.from_

    @property
    def chat_id(self) -> Option[int]:
        """Optional. Message from chat ID. This will be present if the message is sent
        by the bot with the callback button that originated the query."""

        return self.message.map(lambda m: m.v.chat.id)

    @property
    def is_topic_message(self) -> Option[bool]:
        """Optional. True, if the message is a topic message with a name,
        color and icon. This will be present if the message is sent
        by the bot with the callback button that originated the query."""

        return self.message.map(
            lambda m: m.only().map(lambda m: m.is_topic_message.unwrap_or(False)).unwrap_or(False),
        )

    @property
    @unwrapping
    def message_thread_id(self) -> Option[int]:
        """Optional. Unique identifier of the target message thread (for forum supergroups only).
        This will be present if the message is sent
        by the bot with the callback button that originated the query."""

        return self.message.unwrap().only().map(lambda m: m.message_thread_id.unwrap()).cast(Some, Nothing)

    @property
    def message_id(self) -> Option[int]:
        """Optional. Unique message identifier inside this chat. This will be present
        if the message is sent by the bot with the callback button that originated the query.
        """

        return self.message.map(lambda m: m.v.message_id)

    @property
    def chat(self) -> Option[Chat]:
        """Optional. Chat the callback query originated from. This will be present
        if the message is sent by the bot with the callback button that originated the query.
        """

        return self.message.map(lambda m: m.v.chat)

    def decode_callback_data(self) -> Option[dict[str, typing.Any]]:
        if "cached_callback_data" in self.__dict__:
            return self.__dict__["cached_callback_data"]
        data = Nothing()
        with suppress(msgspec.ValidationError):
            data = Some(decoder.decode(self.data.unwrap()))
        self.__dict__["cached_callback_data"] = data
        return data

    @shortcut("answer_callback_query", custom_params={"callback_query_id"})
    async def answer(
        self,
        text: str | None = None,
        callback_query_id: str | None = None,
        show_alert: bool | None = None,
        url: str | None = None,
        cache_time: int | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.answer_callback_query()`, see the [documentation](https://core.telegram.org/bots/api#answercallbackquery)

        Use this method to send answers to callback queries sent from inline keyboards.
        The answer will be displayed to the user as a notification at the top of the
        chat screen or as an alert. On success, True is returned."""

        params = compose_method_params(
            get_params(locals()), self, default_params={("callback_query_id", "id")}
        )
        return await self.ctx_api.answer_callback_query(**params)

    @shortcut(
        "copy_message",
        custom_params={
            "reply_parameters",
            "message_thread_id",
            "chat_id",
            "message_id",
            "from_chat_id",
            "reply_markup",
        },
    )
    async def copy(
        self,
        chat_id: int | str | None = None,
        from_chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        show_caption_above_media: bool | None = None,
        **other: typing.Any,
    ) -> Result[MessageId, APIError]:
        """Shortcut `API.copy_message()`, see the [documentation](https://core.telegram.org/bots/api#copymessage)

        Use this method to copy messages of any kind. Service messages, paid media
        messages, giveaway messages, giveaway winners messages, and invoice
        messages can't be copied. A quiz poll can be copied only if the value of the
        field correct_option_id is known to the bot. The method is analogous to
        the method forwardMessage, but the copied message doesn't have a link to
        the original message. Returns the MessageId of the sent message on success."""

        return await MessageCute.copy(self, **get_params(locals()))  # type: ignore

    @shortcut("delete_message", custom_params={"message_thread_id", "chat_id", "message_id"})
    async def delete(
        self,
        chat_id: int | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.delete_message()`, see the [documentation](https://core.telegram.org/bots/api#deletemessage)

        Use this method to delete a message, including service messages, with the
        following limitations: - A message can only be deleted if it was sent less
        than 48 hours ago. - Service messages about a supergroup, channel, or forum
        topic creation can't be deleted. - A dice message in a private chat can only
        be deleted if it was sent more than 24 hours ago. - Bots can delete outgoing
        messages in private chats, groups, and supergroups. - Bots can delete incoming
        messages in private chats. - Bots granted can_post_messages permissions
        can delete outgoing messages in channels. - If the bot is an administrator
        of a group, it can delete any message there. - If the bot has can_delete_messages
        permission in a supergroup or a channel, it can delete any message there.
        Returns True on success."""

        return await MessageCute.delete(self, **get_params(locals()))  # type: ignore

    @shortcut(
        "edit_message_text",
        executor=execute_method_edit,
        custom_params={"message_thread_id", "link_preview_options"},
    )
    async def edit_text(
        self,
        text: str,
        inline_message_id: str | None = None,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        parse_mode: str | None = None,
        entities: list[MessageEntity] | None = None,
        link_preview_options: LinkPreviewOptions | dict[str, typing.Any] | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        business_connection_id: str | None = None,
        **other: typing.Any,
    ) -> Result[Variative[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_text()`, see the [documentation](https://core.telegram.org/bots/api#editmessagetext)

        Use this method to edit text and game messages. On success, if the edited
        message is not an inline message, the edited Message is returned, otherwise
        True is returned. Note that business messages that were not sent by the bot
        and do not contain an inline keyboard can only be edited within 48 hours from
        the time they were sent.
        :param business_connection_id: Unique identifier of the business connection on behalf of which the messageto be edited was sent.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier forthe target chat or username of the target channel (in the format @channelusername).
        :param message_id: Required if inline_message_id is not specified. Identifier of the messageto edit.

        :param inline_message_id: Required if chat_id and message_id are not specified. Identifier of theinline message.

        :param text: New text of the message, 1-4096 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the message text. See formatting options formore details.

        :param entities: A JSON-serialized list of special entities that appear in message text,which can be specified instead of parse_mode.

        :param link_preview_options: Link preview generation options for the message.

        :param reply_markup: A JSON-serialized object for an inline keyboard."""

        ...

    @shortcut(
        "edit_message_live_location",
        executor=execute_method_edit,
        custom_params={"message_thread_id"},
    )
    async def edit_live_location(
        self,
        latitude: float,
        longitude: float,
        inline_message_id: str | None = None,
        message_thread_id: int | None = None,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        horizontal_accuracy: float | None = None,
        heading: int | None = None,
        proximity_alert_radius: int | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        business_connection_id: str | None = None,
        live_period: int | None = None,
        **other: typing.Any,
    ) -> Result[Variative[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_live_location()`, see the [documentation](https://core.telegram.org/bots/api#editmessagelivelocation)

        Use this method to edit live location messages. A location can be edited
        until its live_period expires or editing is explicitly disabled by a call
        to stopMessageLiveLocation. On success, if the edited message is not an
        inline message, the edited Message is returned, otherwise True is returned.
        :param business_connection_id: Unique identifier of the business connection on behalf of which the messageto be edited was sent.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier forthe target chat or username of the target channel (in the format @channelusername).
        :param message_id: Required if inline_message_id is not specified. Identifier of the messageto edit.

        :param inline_message_id: Required if chat_id and message_id are not specified. Identifier of theinline message.

        :param latitude: Latitude of new location.

        :param longitude: Longitude of new location.

        :param live_period: New period in seconds during which the location can be updated, startingfrom the message send date. If 0x7FFFFFFF is specified, then the locationcan be updated forever. Otherwise, the new value must not exceed the currentlive_period by more than a day, and the live location expiration date mustremain within the next 90 days. If not specified, then live_period remainsunchanged.

        :param horizontal_accuracy: The radius of uncertainty for the location, measured in meters; 0-1500.
        :param heading: Direction in which the user is moving, in degrees. Must be between 1 and 360if specified.

        :param proximity_alert_radius: The maximum distance for proximity alerts about approaching another chatmember, in meters. Must be between 1 and 100000 if specified.

        :param reply_markup: A JSON-serialized object for a new inline keyboard."""

        ...

    @shortcut(
        "edit_message_caption",
        executor=execute_method_edit,
        custom_params={"message_thread_id"},
    )
    async def edit_caption(
        self,
        caption: str | None = None,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        inline_message_id: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        business_connection_id: str | None = None,
        show_caption_above_media: bool | None = None,
        **other: typing.Any,
    ) -> Result[Variative[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_caption()`, see the [documentation](https://core.telegram.org/bots/api#editmessagecaption)

        Use this method to edit captions of messages. On success, if the edited message
        is not an inline message, the edited Message is returned, otherwise True
        is returned. Note that business messages that were not sent by the bot and
        do not contain an inline keyboard can only be edited within 48 hours from
        the time they were sent.
        :param business_connection_id: Unique identifier of the business connection on behalf of which the messageto be edited was sent.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier forthe target chat or username of the target channel (in the format @channelusername).
        :param message_id: Required if inline_message_id is not specified. Identifier of the messageto edit.

        :param inline_message_id: Required if chat_id and message_id are not specified. Identifier of theinline message.

        :param caption: New caption of the message, 0-1024 characters after entities parsing.
        :param parse_mode: Mode for parsing entities in the message caption. See formatting optionsfor more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption,which can be specified instead of parse_mode.

        :param show_caption_above_media: Pass True, if the caption must be shown above the message media. Supportedonly for animation, photo and video messages.

        :param reply_markup: A JSON-serialized object for an inline keyboard."""

        ...

    @shortcut(
        "edit_message_media",
        custom_params={
            "media",
            "type",
            "message_thread_id",
            "caption",
            "parse_mode",
            "caption_entities",
        },
    )
    async def edit_media(
        self,
        media: str | InputFile | InputMedia,
        type: MediaType | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        inline_message_id: str | None = None,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        business_connection_id: str | None = None,
        **other: typing.Any,
    ) -> Result[Variative[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_media()`, see the [documentation](https://core.telegram.org/bots/api#editmessagemedia)

        Use this method to edit animation, audio, document, photo, or video messages.
        If a message is part of a message album, then it can be edited only to an audio
        for audio albums, only to a document for document albums and to a photo or
        a video otherwise. When an inline message is edited, a new file can't be uploaded;
        use a previously uploaded file via its file_id or specify a URL. On success,
        if the edited message is not an inline message, the edited Message is returned,
        otherwise True is returned. Note that business messages that were not sent
        by the bot and do not contain an inline keyboard can only be edited within
        48 hours from the time they were sent.
        :param business_connection_id: Unique identifier of the business connection on behalf of which the messageto be edited was sent.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier forthe target chat or username of the target channel (in the format @channelusername).
        :param message_id: Required if inline_message_id is not specified. Identifier of the messageto edit.

        :param inline_message_id: Required if chat_id and message_id are not specified. Identifier of theinline message.

        :param media: A JSON-serialized object for a new media content of the message.

        :param reply_markup: A JSON-serialized object for a new inline keyboard."""

        return await MessageCute.edit_media(self, **get_params(locals()))  # type: ignore

    @shortcut(
        "edit_message_reply_markup",
        executor=execute_method_edit,
        custom_params={"message_thread_id"},
    )
    async def edit_reply_markup(
        self,
        inline_message_id: str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        chat_id: int | str | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        business_connection_id: str | None = None,
        **other: typing.Any,
    ) -> Result[Variative[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_reply_markup()`, see the [documentation](https://core.telegram.org/bots/api#editmessagereplymarkup)

        Use this method to edit only the reply markup of messages. On success, if
        the edited message is not an inline message, the edited Message is returned,
        otherwise True is returned. Note that business messages that were not sent
        by the bot and do not contain an inline keyboard can only be edited within
        48 hours from the time they were sent.
        :param business_connection_id: Unique identifier of the business connection on behalf of which the messageto be edited was sent.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier forthe target chat or username of the target channel (in the format @channelusername).
        :param message_id: Required if inline_message_id is not specified. Identifier of the messageto edit.

        :param inline_message_id: Required if chat_id and message_id are not specified. Identifier of theinline message.

        :param reply_markup: A JSON-serialized object for an inline keyboard."""

        ...


__all__ = ("CallbackQueryCute",)



import typing
from datetime import datetime

from fntypes.result import Result

from telegrinder.api import API, APIError
from telegrinder.bot.cute_types.base import BaseCute, compose_method_params, shortcut
from telegrinder.model import get_params
from telegrinder.types.objects import ChatMemberUpdated, ChatPermissions, User


async def chat_member_interaction(
    update: BaseCute[typing.Any],
    method_name: str,
    params: dict[str, typing.Any],
) -> Result[typing.Any, APIError]:
    if isinstance(params.get("permissions"), dict):
        params["permissions"] = ChatPermissions(**params["permissions"])
    params = compose_method_params(
        get_params(locals()),
        update,
        default_params={"chat_id", "user_id"},
    )
    return await getattr(update.ctx_api, method_name)(**params)


class ChatMemberShortcuts:
    """Shortcut methods for `ChatMemberUpdatedCute`, `ChatJoinRequestCute` objects."""

    @shortcut("ban_chat_member", executor=chat_member_interaction, custom_params={"chat_id", "user_id"})
    async def ban_chat_member(
        self,
        chat_id: int | str | None = None,
        user_id: int | None = None,
        until_date: datetime | int | None = None,
        revoke_messages: bool | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.ban_chat_member()`, see the [documentation](https://core.telegram.org/bots/api#banchatmember)

        Use this method to ban a user in a group, a supergroup or a channel. In the case
        of supergroups and channels, the user will not be able to return to the chat
        on their own using invite links, etc., unless unbanned first. The bot must
        be an administrator in the chat for this to work and must have the appropriate
        administrator rights. Returns True on success."""

        ...

    @shortcut("unban_chat_member", executor=chat_member_interaction, custom_params={"chat_id", "user_id"})
    async def unban_chat_member(
        self,
        chat_id: int | str | None = None,
        user_id: int | None = None,
        only_if_banned: bool | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.unban_chat_member()`, see the [documentation](https://core.telegram.org/bots/api#unbanchatmember)

        Use this method to unban a previously banned user in a supergroup or channel.
        The user will not return to the group or channel automatically, but will
        be able to join via link, etc. The bot must be an administrator for this to
        work. By default, this method guarantees that after the call the user is
        not a member of the chat, but will be able to join it. So if the user is a member
        of the chat they will also be removed from the chat. If you don't want this,
        use the parameter only_if_banned. Returns True on success."""

        ...

    @shortcut(
        "restrict_chat_member",
        executor=chat_member_interaction,
        custom_params={"permissions", "chat_id", "user_id"},
    )
    async def restrict_chat_member(
        self,
        permissions: ChatPermissions | dict[str, typing.Any],
        chat_id: int | str | None = None,
        user_id: int | None = None,
        use_independent_chat_permissions: bool | None = None,
        until_date: datetime | int | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.restrict_chat_member()`, see the [documentation](https://core.telegram.org/bots/api#restrictchatmember)

        Use this method to restrict a user in a supergroup. The bot must be an administrator
        in the supergroup for this to work and must have the appropriate administrator
        rights. Pass True for all permissions to lift restrictions from a user.
        Returns True on success."""

        ...

    @shortcut("promote_chat_member", executor=chat_member_interaction, custom_params={"chat_id", "user_id"})
    async def promote_chat_member(
        self,
        chat_id: int | str | None = None,
        user_id: int | None = None,
        is_anonymous: bool | None = None,
        can_manage_chat: bool | None = None,
        can_delete_messages: bool | None = None,
        can_manage_video_chats: bool | None = None,
        can_restrict_members: bool | None = None,
        can_promote_members: bool | None = None,
        can_change_info: bool | None = None,
        can_invite_users: bool | None = None,
        can_post_stories: bool | None = None,
        can_edit_stories: bool | None = None,
        can_delete_stories: bool | None = None,
        can_post_messages: bool | None = None,
        can_edit_messages: bool | None = None,
        can_pin_messages: bool | None = None,
        can_manage_topics: bool | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.promote_chat_member()`, see the [documentation](https://core.telegram.org/bots/api#promotechatmember)

        Use this method to promote or demote a user in a supergroup or a channel. The
        bot must be an administrator in the chat for this to work and must have the
        appropriate administrator rights. Pass False for all boolean parameters
        to demote a user. Returns True on success."""

        ...

    @shortcut(
        "set_chat_administrator_custom_title",
        executor=chat_member_interaction,
        custom_params={"chat_id", "user_id"},
    )
    async def set_chat_administrator_custom_title(
        self,
        custom_title: str,
        chat_id: int | str | None = None,
        user_id: int | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.set_chat_administrator_custom_title()`, see the [documentation](https://core.telegram.org/bots/api#setchatadministratorcustomtitle)

        Use this method to set a custom title for an administrator in a supergroup
        promoted by the bot. Returns True on success."""

        ...


class ChatMemberUpdatedCute(
    BaseCute[ChatMemberUpdated], ChatMemberUpdated, ChatMemberShortcuts, kw_only=True
):
    api: API

    @property
    def from_user(self) -> User:
        return self.from_

    @property
    def user_id(self) -> int:
        return self.from_user.id


__all__ = (
    "ChatMemberShortcuts",
    "ChatMemberUpdatedCute",
)



import pathlib
import typing
from functools import cached_property

from envparse import env

from .error import InvalidTokenError


class Token(str):
    def __new__(cls, token: str) -> typing.Self:
        if token.count(":") != 1 or not token.split(":")[0].isdigit():
            raise InvalidTokenError("Invalid token, it should look like this '123:ABC'.")
        return super().__new__(cls, token)

    def __repr__(self) -> str:
        return f"<Token: {self.bot_id}:{''.join(self.split(':')[-1])[:6]}...>"

    @classmethod
    def from_env(
        cls,
        var_name: str = "BOT_TOKEN",
        *,
        is_read: bool = False,
        path_to_envfile: str | pathlib.Path | None = None,
    ) -> typing.Self:
        if not is_read:
            env.read_envfile(path_to_envfile)
        return cls(env.str(var_name))

    @cached_property
    def bot_id(self) -> int:
        return int(self.split(":")[0])


__all__ = ("Token",)



from telegrinder.bot.rules.abc import ABCRule, AndRule, NotRule, OrRule
from telegrinder.bot.rules.callback_data import (
    CallbackDataEq,
    CallbackDataJsonEq,
    CallbackDataJsonModel,
    CallbackDataMap,
    CallbackDataMarkup,
    CallbackQueryDataRule,
    CallbackQueryRule,
    HasData,
)
from telegrinder.bot.rules.chat_join import (
    ChatJoinRequestRule,
    HasInviteLink,
    InviteLinkByCreator,
    InviteLinkName,
)
from telegrinder.bot.rules.command import Argument, Command
from telegrinder.bot.rules.enum_text import EnumTextRule
from telegrinder.bot.rules.func import FuncRule
from telegrinder.bot.rules.fuzzy import FuzzyText
from telegrinder.bot.rules.inline import (
    HasLocation,
    InlineQueryChatType,
    InlineQueryMarkup,
    InlineQueryRule,
    InlineQueryText,
)
from telegrinder.bot.rules.integer import IntegerInRange, IsInteger
from telegrinder.bot.rules.is_from import (
    IsBot,
    IsChat,
    IsChatId,
    IsDice,
    IsDiceEmoji,
    IsForum,
    IsForward,
    IsForwardType,
    IsGroup,
    IsLanguageCode,
    IsPremium,
    IsPrivate,
    IsReply,
    IsSuperGroup,
    IsUser,
    IsUserId,
)
from telegrinder.bot.rules.markup import Markup
from telegrinder.bot.rules.mention import HasMention
from telegrinder.bot.rules.message import MessageRule
from telegrinder.bot.rules.message_entities import HasEntities, MessageEntities
from telegrinder.bot.rules.node import NodeRule
from telegrinder.bot.rules.regex import Regex
from telegrinder.bot.rules.rule_enum import RuleEnum
from telegrinder.bot.rules.start import StartCommand
from telegrinder.bot.rules.state import State, StateMeta
from telegrinder.bot.rules.text import HasText, Text
from telegrinder.bot.rules.update import IsUpdateType

__all__ = (
    "ABCRule",
    "AndRule",
    "Argument",
    "CallbackDataEq",
    "CallbackDataJsonEq",
    "CallbackDataJsonModel",
    "CallbackDataMap",
    "CallbackDataMarkup",
    "CallbackQueryDataRule",
    "CallbackQueryRule",
    "ChatJoinRequestRule",
    "Command",
    "EnumTextRule",
    "FuncRule",
    "FuzzyText",
    "HasData",
    "HasEntities",
    "HasInviteLink",
    "HasLocation",
    "HasMention",
    "HasText",
    "InlineQueryChatType",
    "InlineQueryMarkup",
    "InlineQueryRule",
    "InlineQueryText",
    "IntegerInRange",
    "InviteLinkByCreator",
    "InviteLinkName",
    "IsBot",
    "IsChat",
    "IsChatId",
    "IsDice",
    "IsDiceEmoji",
    "IsForum",
    "IsForward",
    "IsForwardType",
    "IsGroup",
    "IsInteger",
    "IsLanguageCode",
    "IsPremium",
    "IsPrivate",
    "IsReply",
    "IsSuperGroup",
    "IsUpdateType",
    "IsUser",
    "IsUserId",
    "Markup",
    "MessageEntities",
    "MessageRule",
    "NodeRule",
    "NotRule",
    "OrRule",
    "Regex",
    "RuleEnum",
    "StartCommand",
    "State",
    "StateMeta",
    "Text",
)



import typing

from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.return_manager.abc import BaseReturnManager, register_manager
from telegrinder.tools.formatting import HTMLFormatter


class MessageReturnManager(BaseReturnManager[MessageCute]):
    @register_manager(str)
    @staticmethod
    async def str_manager(value: str, event: MessageCute, ctx: Context) -> None:
        await event.answer(value)

    @register_manager(list[str] | tuple[str, ...])
    @staticmethod
    async def seq_manager(
        value: list[str] | tuple[str, ...],
        event: MessageCute,
        ctx: Context,
    ) -> None:
        for message in value:
            await event.answer(message)

    @register_manager(dict[str, typing.Any])
    @staticmethod
    async def dict_manager(value: dict[str, typing.Any], event: MessageCute, ctx: Context) -> None:
        await event.answer(**value)

    @register_manager(HTMLFormatter)
    @staticmethod
    async def htmlformatter_manager(value: HTMLFormatter, event: MessageCute, ctx: Context) -> None:
        await event.answer(value, parse_mode=HTMLFormatter.PARSE_MODE)


__all__ = ("MessageReturnManager",)



import enum
import inspect
import types
import typing
from functools import wraps

if typing.TYPE_CHECKING:
    from telegrinder.bot.rules.abc import ABCRule
    from telegrinder.node.polymorphic import Polymorphic

    T = typing.TypeVar("T", bound=ABCRule)
    F = typing.TypeVar(
        "F",
        bound=typing.Callable[typing.Concatenate[typing.Callable[..., typing.Any], ...], typing.Any],
    )

Impl: typing.TypeAlias = type[classmethod]
FuncType: typing.TypeAlias = types.FunctionType | typing.Callable[..., typing.Any]

TRANSLATIONS_KEY: typing.Final[str] = "_translations"
IMPL_MARK: typing.Final[str] = "_is_impl"


def cache_magic_value(mark_key: str, /):
    def inner(func: "F") -> "F":
        @wraps(func)
        def wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
            if mark_key not in args[0].__dict__:
                args[0].__dict__[mark_key] = func(*args, **kwargs)
            return args[0].__dict__[mark_key]

        return wrapper  # type: ignore

    return inner


def resolve_arg_names(func: FuncType, start_idx: int = 1) -> tuple[str, ...]:
    return func.__code__.co_varnames[start_idx : func.__code__.co_argcount]


@cache_magic_value("__default_args__")
def get_default_args(func: FuncType) -> dict[str, typing.Any]:
    fspec = inspect.getfullargspec(func)
    if not fspec.defaults:
        return {}
    return dict(zip(fspec.args[-len(fspec.defaults) :], fspec.defaults))


def get_annotations(func: FuncType, *, return_type: bool = False) -> dict[str, typing.Any]:
    annotations = func.__annotations__
    if not return_type and "return" in func.__annotations__:
        annotations.pop("return")
    return annotations


def to_str(s: str | enum.Enum) -> str:
    if isinstance(s, enum.Enum):
        return str(s.value)
    return s


@typing.overload
def magic_bundle(handler: FuncType, kw: dict[str, typing.Any]) -> dict[str, typing.Any]: ...


@typing.overload
def magic_bundle(handler: FuncType, kw: dict[enum.Enum, typing.Any]) -> dict[str, typing.Any]: ...


@typing.overload
def magic_bundle(
    handler: FuncType,
    kw: dict[str, typing.Any],
    *,
    start_idx: int = 1,
    bundle_ctx: bool = True,
) -> dict[str, typing.Any]: ...


@typing.overload
def magic_bundle(
    handler: FuncType,
    kw: dict[enum.Enum, typing.Any],
    *,
    start_idx: int = 1,
    bundle_ctx: bool = True,
) -> dict[str, typing.Any]: ...


@typing.overload
def magic_bundle(
    handler: FuncType,
    kw: dict[type, typing.Any],
    *,
    typebundle: typing.Literal[True] = True,
) -> dict[str, typing.Any]: ...


def magic_bundle(
    handler: FuncType,
    kw: dict[typing.Any, typing.Any],
    *,
    start_idx: int = 1,
    bundle_ctx: bool = True,
    typebundle: bool = False,
) -> dict[str, typing.Any]:
    if typebundle:
        types = get_annotations(handler, return_type=False)
        bundle: dict[str, typing.Any] = {}
        for name, type in types.items():
            bundle[name] = kw[type]
        return bundle

    names = resolve_arg_names(handler, start_idx=start_idx)
    args = get_default_args(handler)
    args.update({to_str(k): v for k, v in kw.items() if to_str(k) in names})
    if "ctx" in names and bundle_ctx:
        args["ctx"] = kw
    return args


def get_cached_translation(rule: "T", locale: str) -> "T | None":
    return getattr(rule, TRANSLATIONS_KEY, {}).get(locale)


def cache_translation(base_rule: "T", locale: str, translated_rule: "T") -> None:
    translations = getattr(base_rule, TRANSLATIONS_KEY, {})
    translations[locale] = translated_rule
    setattr(base_rule, TRANSLATIONS_KEY, translations)


@typing.cast(typing.Callable[..., Impl], lambda f: f)
def impl(method: typing.Callable[..., typing.Any]):
    setattr(method, IMPL_MARK, True)
    return classmethod(method)


def get_impls(cls: type["Polymorphic"]) -> list[typing.Callable[..., typing.Any]]:
    moprh_impls = getattr(cls, "__morph_impls__", None)
    if moprh_impls is not None:
        return moprh_impls
    impls = [
        func.__func__
        for func in vars(cls).values()
        if isinstance(func, classmethod) and getattr(func.__func__, IMPL_MARK, False)
    ]
    setattr(cls, "__morph_impls__", impls)
    return impls


__all__ = (
    "TRANSLATIONS_KEY",
    "cache_magic_value",
    "cache_translation",
    "get_annotations",
    "get_cached_translation",
    "get_default_args",
    "get_default_args",
    "get_impls",
    "impl",
    "magic_bundle",
    "resolve_arg_names",
    "to_str",
)
